# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import signal
import sys
import time
from collections import namedtuple
from typing import Tuple

import tensorflow as tf
from absl import logging
from tensorflow.python.distribute import distribution_strategy_context, reduce_util
from tensorflow.python.framework import dtypes
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.training.experimental.loss_scale import DynamicLossScale, _assign_if_finite, _op_in_graph_mode
from tensorflow.python.util import nest

from neurst.criterions import Criterion
from neurst.data.datasets import Dataset
from neurst.data.datasets.multiple_dataset import MultipleDataset
from neurst.training import distribution_utils
from neurst.utils import compat
from neurst.utils.checkpoints import AverageCheckpointSaver, KeepBestCheckpointSaver
from neurst.utils.configurable import ModelConfigs
from neurst.utils.misc import DummyContextManager, to_numpy_or_python_type

GPU_EFFICIENT_LEVEL = namedtuple("gpu_efficient_level",
                                 "LEVEL0 LEVEL1 LEVEL2 LEVEL3 LEVEL4 LEVEL5")(0, 1, 2, 3, 4, 5)

EFFICIENT_MULTIPLIER = {
    GPU_EFFICIENT_LEVEL.LEVEL0: 8,
    GPU_EFFICIENT_LEVEL.LEVEL1: 8,
    GPU_EFFICIENT_LEVEL.LEVEL2: 16,
    GPU_EFFICIENT_LEVEL.LEVEL3: 32,
    GPU_EFFICIENT_LEVEL.LEVEL4: 64,
    GPU_EFFICIENT_LEVEL.LEVEL5: 8
}


def minimal_multiple(val, factor):
    if val % factor == 0:
        return val
    return int((val // factor + 1) * factor)


def maximum_lower_multiple(val, factor):
    multiple = minimal_multiple(val, factor)
    if (multiple - val) // factor > 0.5:
        cand = multiple - factor
        if cand <= 0:
            return multiple
        return cand
    return multiple


def startup_env(dtype="float16",
                enable_check_numerics=False,
                enable_xla=False):
    """ Startup program environments. """
    if dtype not in ["float32", "float16"]:
        raise ValueError(f"Not supported dtype={dtype} (now only accept float32 and float16)")
    if dtype == "float16":
        logging.info("Using float16 as computation dtype.")
        from tensorflow.keras.mixed_precision import experimental as mixed_precision
        policy = mixed_precision.Policy("mixed_float16", loss_scale="dynamic")
        mixed_precision.set_policy(policy)
        compat.register_computation_dtype("float16", -6.e4)
    if enable_check_numerics:
        logging.info("Enable checking numerics.")
        tf.debugging.enable_check_numerics()
    if enable_xla:
        tf.config.optimizer.set_jit(True)
        # it causes OOM and performance reression
        tf.config.optimizer.set_experimental_options({"pin_to_host_optimization": False})


def is_third_party_allreduce(strategy):
    return strategy in ["byteps", "horovod"]


def handle_distribution_strategy(distribution_strategy):
    """ Create distribution strategy. """
    strategy = None
    if distribution_strategy:
        strategy = distribution_strategy
        if isinstance(distribution_strategy, dict):
            strategy = distribution_strategy.get("distribution_strategy", None)
        if isinstance(distribution_strategy, str):
            strategy = distribution_strategy.lower()
        if is_third_party_allreduce(strategy):
            if strategy == "horovod":
                import horovod.tensorflow.keras as hvd
            else:
                import byteps.tensorflow.keras as hvd
            logging.info("import {} as hvd backend.".format(strategy))
            hvd.init()
            # Horovod: pin GPU to be used to process local rank (one GPU per process)
            gpus = tf.config.experimental.list_physical_devices('GPU')
            for gpu in gpus:
                tf.config.experimental.set_memory_growth(gpu, True)
            if gpus:
                tf.config.experimental.set_visible_devices(gpus[hvd.local_rank()], 'GPU')
            compat.register_distributed_worker_setting(hvd.rank(), hvd.size(), strategy)
            if hvd.rank() != 0:
                logging.set_verbosity(logging.ERROR)
        else:
            if isinstance(distribution_strategy, str):
                strategy = distribution_utils.get_distribution_strategy(
                    distribution_strategy=distribution_strategy)
            elif isinstance(distribution_strategy, dict):
                strategy = distribution_utils.get_distribution_strategy(**distribution_strategy)

    if strategy is None:
        logging.info("No distribution strategy was used.")
    else:
        try:
            logging.info("Using distribution strategy: {} with num_replicas_in_sync={}".format(
                strategy, strategy.num_replicas_in_sync))
        except Exception:
            pass
    return strategy


def get_strategy_scope(strategy):
    """ Returns the distributed strategy context. """
    try:
        return strategy.scope()
    except AttributeError:
        return DummyContextManager()


def get_num_replicas_in_sync(strategy):
    """ Returns the number of replicas in sync. """
    try:
        return strategy.num_replicas_in_sync
    except AttributeError:
        return 1


_SINGLE_DS_NAME = ""


def build_datasets(mode,
                   strategy,
                   custom_dataset: Dataset,
                   task,
                   args=None):
    """ Builds datasets and returns datasets. """
    if isinstance(custom_dataset, MultipleDataset):
        return {
            name: task.create_and_batch_tfds(
                d, mode, num_replicas_in_sync=get_num_replicas_in_sync(strategy), args=args)
            for name, d in custom_dataset.datasets.items()
        }
    return task.create_and_batch_tfds(
        custom_dataset, mode, num_replicas_in_sync=get_num_replicas_in_sync(strategy), args=args)


def maybe_distribution_dataset(strategy, dataset):
    try:
        return strategy.experimental_distribute_dataset(dataset)
    except AttributeError:
        return dataset


def make_predictions(strategy,
                     model: tf.keras.models.Model,
                     tfds,
                     custom_dataset,
                     map_func=None):
    """ Makes predictions.
        The `tf_ds` are created by `build_datasets` fn.

    Args:
        strategy: The distritbution strategy.
        model: The keras model.
        tfds: A dataset or a list of datasets returned by `build_datasets`.
        custom_dataset: A Dataset object.
        map_func: A callable function that tasks custom dataset and eval result as inputs
            and converters each eval result.

    Returns: A dict of evaluation results for each dataset
        or the evaluation result for single dataset.
    """
    if not isinstance(custom_dataset, MultipleDataset):
        tfds = {_SINGLE_DS_NAME: tfds}
    results = {}
    with get_strategy_scope(strategy):
        predict_fn = model.make_predict_function()
        for name, ds in tfds.items():
            assert isinstance(ds, tf.data.Dataset), (
                "Unsupported type of dataset({}): {}".format(ds, type(ds)))
            iterator = iter(maybe_distribution_dataset(strategy, ds.prefetch(
                tf.data.experimental.AUTOTUNE)))
            this_results = []
            while True:
                try:
                    this_results.append(predict_fn(iterator))
                except (StopIteration, tf.errors.OutOfRangeError):
                    break
            if map_func is None:
                results[name] = this_results
            else:
                results[name] = map_func(this_results)
        if not isinstance(custom_dataset, MultipleDataset):
            results = results[_SINGLE_DS_NAME]
    return results


def reduce_eval_results(criterion: Criterion,
                        custom_dataset: Dataset,
                        eval_results) -> Tuple[dict, dict, dict]:
    """

    Args:
        criterion: A criterion instance.
        custom_dataset: The custom dataset object.
        eval_results: The prediction results from `make_predictions`.

    Returns: A tuple of dicts of evaluation results:
        - result dict of each dataset
        - averaged result
        - mixed result
    """
    if isinstance(custom_dataset, MultipleDataset):
        res = {}
        avg_res = {}
        mixed_data = []
        for k, v in eval_results.items():
            res[k] = criterion.reduce_metrics(v)
            mixed_data.extend(v)
            for kk, vv in res[k].items():
                if kk not in avg_res:
                    avg_res[kk] = 0.
                avg_res[kk] += vv * custom_dataset.sample_weights[k]
        mixed_res = criterion.reduce_metrics(mixed_data)
    else:
        res = avg_res = mixed_res = criterion.reduce_metrics(eval_results)
    return (to_numpy_or_python_type(res), to_numpy_or_python_type(avg_res),
            to_numpy_or_python_type(mixed_res))


class TrainingStatusRecorder(object):
    """ Manage the training status with the best metrics. """

    def __init__(self,
                 model,
                 task,
                 metric,
                 estop_patience=None,
                 best_checkpoint_path=None,
                 auto_average_checkpoints=True,
                 best_avg_checkpoint_path=None,
                 top_checkpoints_to_keep=0):
        """ Initializes manager for arbitrary evaluation strategies.

        Args:
            model: The custom keras model (inherent BaseModel).
            task: The custom task.
            metric: The evaluation metric object.
            estop_patience: An integer, the training process will automatically shut down until the program
                fail to acquire a better metric score anymore if `early_stop_patience` greater than 0.
            best_checkpoint_path: The path for checkpoints with best metric scores if provided,
                otherwise, default \"`model_dir`_best\" will be used.
            best_avg_checkpoint_path: The path to saving the averaged checkpoints.
            auto_average_checkpoints: A boolean, whether to do checkpoint average on all model weights.
                An extra directory for averaged weights will be created. It is only available when
                `eval_best_checkpoint_path` is provided.
            top_checkpoints_to_keep: An integer, the maximum number of checkpoints to be saved
                (`max_to_keep` for checkpoint manager), and the number of latest checkpoints to be averaged
                if `eval_auto_average_checkpoints` is True. If <= 0, no more checkpoints will be saved.
        """
        self._model = model
        self._task = task
        self._metric = metric
        self._estop_patience = estop_patience
        self._best_checkpoint_path = best_checkpoint_path
        self._auto_average_checkpoints = auto_average_checkpoints
        self._best_avg_checkpoint_path = best_avg_checkpoint_path
        self._top_checkpoints_to_keep = top_checkpoints_to_keep
        self._keep_best_ckpt_saver = None
        self._average_ckpt_saver = None
        if self._top_checkpoints_to_keep and self._top_checkpoints_to_keep > 0:
            self._keep_best_ckpt_saver = KeepBestCheckpointSaver(
                model=self._model,
                directory=self._best_checkpoint_path,
                metric=self._metric,
                max_to_keep=self._top_checkpoints_to_keep)
            ModelConfigs.dump(self._task.model_configs(self._model), self._keep_best_ckpt_saver.directory)
            if self._auto_average_checkpoints:
                self._average_ckpt_saver = AverageCheckpointSaver(
                    model=self._model,
                    directory=self._best_avg_checkpoint_path,
                    metric=self._metric,
                    max_to_keep=self._top_checkpoints_to_keep)
                ModelConfigs.dump(self._task.model_configs(self._model), self._average_ckpt_saver.directory)
        self._best_metric_result = None
        self._bad_count = 0

    @property
    def best(self):
        return self._best_metric_result

    def record(self, step, metric_result):
        """ Records the metrics and keep the best. """
        metric_result = to_numpy_or_python_type(metric_result)
        if (self._best_metric_result is None
            or self._metric.greater_or_eq(metric_result, self._best_metric_result)):
            self._bad_count = 0
            self._best_metric_result = metric_result
        else:
            self._bad_count += 1

        # re-save the best checkpoint
        if self._keep_best_ckpt_saver is not None:
            start_time = time.time()
            stat = self._keep_best_ckpt_saver.save(step, metric_result)
            logging.info("Checking the best checkpoints kept and %s. Elapsed %.2fs",
                         "a new checkpoint was saved" if stat else "no checkpoint was saved.",
                         time.time() - start_time)
        if self._average_ckpt_saver is not None:
            start_time = time.time()
            stat = self._average_ckpt_saver.save(step, metric_result)
            if stat:
                logging.info("An averaged checkpoint was saved. Elapsed %.2fs", time.time() - start_time)

        if self._estop_patience is not None:
            logging.info(f"Evaluating {self._metric.flag} at step={step} with bad count={self._bad_count} "
                         f"(early_stop_patience={self._estop_patience}).")
        if self._estop_patience and self._bad_count >= self._estop_patience > 0:
            logging.info("Hit maximum patience! Early Stop!!!")

            # kill self and exit with code=0
            def handler(*args):
                sys.exit(0)

            # register for signal
            signal.signal(signal.SIGUSR1, handler)
            os.kill(os.getpid(), signal.SIGUSR1)


# (TensorFlow 2.3!) there is a bug on math_ops.reduce_all under Horovod+fp16+XLA
def _refacor_is_all_finite(grads):
    """Returns a scalar boolean tensor indicating if all gradients are finite."""
    is_finite_per_grad = []
    for g in grads:
        if g is None:
            continue
        # is_not_finite = tf.logical_not(tf.math.is_finite(g))
        # reduced_is_finite = tf.logical_not(tf.reduce_any(is_not_finite))

        is_finite = tf.math.is_finite(g)
        reduced_is_finite = tf.equal(tf.size(g, out_type=tf.int64),
                                     tf.math.reduce_sum(tf.cast(is_finite, tf.int64)))
        is_finite_per_grad.append(reduced_is_finite)
    return tf.math.reduce_all(is_finite_per_grad)


class RevisedDynamicLossScale(DynamicLossScale):
    def update(self, grads):
        """Updates loss scale based on if gradients are finite in current step."""
        grads = nest.flatten(grads)
        if distribution_strategy_context.has_strategy():
            distribution = distribution_strategy_context.get_cross_replica_context()

            def get_is_finite(grads):
                is_finite = _refacor_is_all_finite(grads)
                # We cast to float, because we cannot reduce booleans with
                # DistributionStrategy.
                return tf.cast(is_finite, dtypes.float32)

            is_finite_float = distribution.extended.call_for_each_replica(
                get_is_finite, args=(grads,))
            reduced_is_finite_float = distribution.reduce(reduce_util.ReduceOp.SUM,
                                                          is_finite_float, axis=None)
            is_finite = tf.equal(reduced_is_finite_float,
                                 distribution.num_replicas_in_sync)
        else:
            is_finite = _refacor_is_all_finite(grads)

        def update_if_finite_grads():
            """Update assuming the gradients are finite."""

            def incr_loss_scale():
                new_loss_scale = self._current_loss_scale * self._multiplier
                return control_flow_ops.group(
                    _assign_if_finite(self._current_loss_scale, new_loss_scale),
                    self._num_good_steps.assign(0))

            return control_flow_ops.cond(
                self._num_good_steps + 1 >= self._increment_period,
                incr_loss_scale, lambda: _op_in_graph_mode(
                    self._num_good_steps.assign_add(1)))

        def update_if_not_finite_grads():
            """Update assuming the gradients are nonfinite."""

            new_loss_scale = tf.math.maximum(
                self._current_loss_scale / self._multiplier, 1)
            return control_flow_ops.group(
                self._num_good_steps.assign(0),
                self._current_loss_scale.assign(new_loss_scale))

        update_op = control_flow_ops.cond(is_finite, update_if_finite_grads,
                                          update_if_not_finite_grads)
        should_apply_gradients = is_finite
        return update_op, should_apply_gradients



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import numpy
import tensorflow as tf
from absl import logging

from neurst.models.transformer import Transformer
from neurst.utils.converters import Converter, register_converter


@register_converter
class FairseqTransformer(Converter):
    """ Converts from the fairseq's Transformer model.
    """

    @staticmethod
    def convert_model_config(path):
        import torch
        with tf.io.gfile.GFile(path, "rb") as fp:
            cp = torch.load(fp, map_location=torch.device('cpu'))
        args = cp["args"].__dict__
        return {
            "model.class": Transformer.__name__,
            "model.params": {
                "modality.share_source_target_embedding": args["share_all_embeddings"],
                "modality.share_embedding_and_softmax_weights": args["share_decoder_input_output_embed"],
                "modality.source.dim": args["encoder_embed_dim"],
                "modality.target.dim": args["decoder_embed_dim"],
                "modality.source.timing": {"timing": "sinusoids",
                                           "max_positions": args["max_source_positions"] - 16,
                                           "sinusoids_as_variable": True},
                "modality.target.timing": {"timing": "sinusoids",
                                           "max_positions": args["max_target_positions"] - 16,
                                           "sinusoids_as_variable": True},
                "encoder.num_layers": args["encoder_layers"],
                "encoder.hidden_size": args["encoder_embed_dim"],
                "encoder.num_attention_heads": args["encoder_attention_heads"],
                "encoder.filter_size": args["encoder_ffn_embed_dim"],
                "encoder.attention_dropout_rate": args["dropout"],
                "encoder.attention_type": "dot_product",
                "encoder.ffn_activation": "relu",
                "encoder.ffn_dropout_rate": args["dropout"],
                "encoder.post_normalize": (not args["encoder_normalize_before"]),
                "encoder.layer_postprocess_dropout_rate": args["dropout"],
                "decoder.num_layers": args["decoder_layers"],
                "decoder.hidden_size": args["encoder_embed_dim"],
                "decoder.num_attention_heads": args["decoder_attention_heads"],
                "decoder.filter_size": args["decoder_ffn_embed_dim"],
                "decoder.attention_dropout_rate": args["dropout"],
                "decoder.attention_type": "dot_product",
                "decoder.ffn_activation": "relu",
                "decoder.ffn_dropout_rate": args["dropout"],
                "decoder.post_normalize": (not args["decoder_normalize_before"]),
                "decoder.layer_postprocess_dropout_rate": args["dropout"]
            }
        }

    @staticmethod
    def convert_task_config(path):
        # return {
        #     "task.params": {
        #         "target_begin_of_sentence": "eos"
        #     }
        # }
        return None

    @staticmethod
    def convert_checkpoint(path, save_path):
        import torch
        with tf.io.gfile.GFile(path, "rb") as fp:
            pyvars = torch.load(fp, map_location=torch.device('cpu'))["model"]
        cfgs: dict = FairseqTransformer.convert_model_config(path)["model.params"]
        pyvar_names = [x for x in pyvars]
        new_var_dict = {}

        def get_pyvar(name):
            pyvar_names.remove(name)
            return pyvars[name].detach().numpy()

        def reform_emb(emb):
            # vocab: 0-4  --  bos, pad, eos, unk
            print(emb.shape)
            weight = emb[4:]
            bos = emb[0:1]
            eos = emb[2:3]
            unk = emb[3:4]
            return numpy.concatenate([weight, unk, bos, eos], axis=0)

        def get_position_emb(num, dim):
            from fairseq.modules.sinusoidal_positional_embedding import SinusoidalPositionalEmbedding
            return SinusoidalPositionalEmbedding.get_embedding(num + 2, dim, 1).detach().numpy()[2:]

        # embedding table
        tf_prefix = "SequenceToSequence"
        if cfgs["modality.share_source_target_embedding"]:
            _name = "shared_symbol_modality"
            if not cfgs["modality.share_embedding_and_softmax_weights"]:
                raise NotImplementedError
            emb_table = reform_emb(get_pyvar("encoder.embed_tokens.weight"))
            new_var_dict[f"{tf_prefix}/{_name}_posenc_wrapper/{_name}/shared/weights"] = emb_table
            new_var_dict[f"{tf_prefix}/{_name}_posenc_wrapper/position_embeddings/weights"] = get_position_emb(
                cfgs["modality.target.timing"]["max_positions"], cfgs["modality.target.dim"])
            new_var_dict[f"{tf_prefix}/{_name}_posenc_wrapper/{_name}/shared/bias"] = numpy.zeros(
                [emb_table.shape[0], ], dtype=float)
        else:
            _src_name = "input_symbol_modality"
            new_var_dict[f"{tf_prefix}/{_src_name}_posenc_wrapper/{_src_name}/emb/weights"] = reform_emb(
                get_pyvar("encoder.embed_tokens.weight"))
            _trg_name = "target_symbol_modality"
            _emb_name = "emb"
            if cfgs["modality.share_embedding_and_softmax_weights"]:
                _emb_name = "shared"
            trg_emb_table = reform_emb(get_pyvar("decoder.embed_tokens.weight"))
            new_var_dict[f"{tf_prefix}/{_trg_name}_posenc_wrapper/{_trg_name}/{_emb_name}/weights"] = trg_emb_table
            new_var_dict[f"{tf_prefix}/{_src_name}_posenc_wrapper/position_embeddings/weights"] = get_position_emb(
                cfgs["modality.source.timing"]["max_positions"], cfgs["modality.source.dim"])
            new_var_dict[f"{tf_prefix}/{_trg_name}_posenc_wrapper/position_embeddings/weights"] = get_position_emb(
                cfgs["modality.target.timing"]["max_positions"], cfgs["modality.target.dim"])
            new_var_dict[f"{tf_prefix}/{_trg_name}_posenc_wrapper/{_trg_name}/{_emb_name}/bias"] = numpy.zeros(
                [trg_emb_table.shape[0], ], dtype=float)

        # encoder stack
        for i in range(cfgs["encoder.num_layers"]):
            _name = f"{tf_prefix}/TransformerEncoder/layer_{i}/"
            _pyname = f"encoder.layers.{i}."
            _selfatt_name = _name + "self_attention_prepost_wrapper/self_attention/"
            new_var_dict[_selfatt_name + "qkv_transform/kernel"] = get_pyvar(
                _pyname + "self_attn.in_proj_weight").transpose()
            new_var_dict[_selfatt_name + "qkv_transform/bias"] = get_pyvar(_pyname + "self_attn.in_proj_bias")
            new_var_dict[_selfatt_name + "output_transform/kernel"] = get_pyvar(
                _pyname + "self_attn.out_proj.weight").transpose()
            new_var_dict[_selfatt_name + "output_transform/bias"] = get_pyvar(_pyname + "self_attn.out_proj.bias")
            new_var_dict[_name + "self_attention_prepost_wrapper/ln/gamma"] = get_pyvar(
                _pyname + "self_attn_layer_norm.weight")
            new_var_dict[_name + "self_attention_prepost_wrapper/ln/beta"] = get_pyvar(
                _pyname + "self_attn_layer_norm.bias")
            _ffn_name = _name + "ffn_prepost_wrapper/ffn/"
            new_var_dict[_ffn_name + "dense1/kernel"] = get_pyvar(_pyname + "fc1.weight").transpose()
            new_var_dict[_ffn_name + "dense1/bias"] = get_pyvar(_pyname + "fc1.bias")
            new_var_dict[_ffn_name + "dense2/kernel"] = get_pyvar(_pyname + "fc2.weight").transpose()
            new_var_dict[_ffn_name + "dense2/bias"] = get_pyvar(_pyname + "fc2.bias")
            new_var_dict[_name + "ffn_prepost_wrapper/ln/gamma"] = get_pyvar(_pyname + "final_layer_norm.weight")
            new_var_dict[_name + "ffn_prepost_wrapper/ln/beta"] = get_pyvar(_pyname + "final_layer_norm.bias")
        if not cfgs["encoder.post_normalize"]:
            new_var_dict[f"{tf_prefix}/TransformerEncoder/output_ln/gamma"] = get_pyvar("encoder.layer_norm.weight")
            new_var_dict[f"{tf_prefix}/TransformerEncoder/output_ln/beta"] = get_pyvar("encoder.layer_norm.bias")

        # decoder stack
        for i in range(cfgs["decoder.num_layers"]):
            _name = f"{tf_prefix}/TransformerDecoder/layer_{i}/"
            _pyname = f"decoder.layers.{i}."
            _selfatt_name = _name + "self_attention_prepost_wrapper/self_attention/"
            new_var_dict[_selfatt_name + "qkv_transform/kernel"] = get_pyvar(
                _pyname + "self_attn.in_proj_weight").transpose()
            new_var_dict[_selfatt_name + "qkv_transform/bias"] = get_pyvar(_pyname + "self_attn.in_proj_bias")
            new_var_dict[_selfatt_name + "output_transform/kernel"] = get_pyvar(
                _pyname + "self_attn.out_proj.weight").transpose()
            new_var_dict[_selfatt_name + "output_transform/bias"] = get_pyvar(_pyname + "self_attn.out_proj.bias")
            new_var_dict[_name + "self_attention_prepost_wrapper/ln/gamma"] = get_pyvar(
                _pyname + "self_attn_layer_norm.weight")
            new_var_dict[_name + "self_attention_prepost_wrapper/ln/beta"] = get_pyvar(
                _pyname + "self_attn_layer_norm.bias")

            _encatt_name = _name + "encdec_attention_prepost_wrapper/encdec_attention/"
            qkv_w = get_pyvar(_pyname + "encoder_attn.in_proj_weight")
            qkv_b = get_pyvar(_pyname + "encoder_attn.in_proj_bias")
            new_var_dict[_encatt_name + "kv_transform/kernel"] = qkv_w[cfgs["decoder.hidden_size"]:].transpose()
            new_var_dict[_encatt_name + "kv_transform/bias"] = qkv_b[cfgs["decoder.hidden_size"]:]
            new_var_dict[_encatt_name + "q_transform/kernel"] = qkv_w[:cfgs["decoder.hidden_size"]].transpose()
            new_var_dict[_encatt_name + "q_transform/bias"] = qkv_b[:cfgs["decoder.hidden_size"]]
            new_var_dict[_encatt_name + "output_transform/kernel"] = get_pyvar(
                _pyname + "encoder_attn.out_proj.weight").transpose()
            new_var_dict[_encatt_name + "output_transform/bias"] = get_pyvar(_pyname + "encoder_attn.out_proj.bias")
            new_var_dict[_name + "encdec_attention_prepost_wrapper/ln/gamma"] = get_pyvar(
                _pyname + "encoder_attn_layer_norm.weight")
            new_var_dict[_name + "encdec_attention_prepost_wrapper/ln/beta"] = get_pyvar(
                _pyname + "encoder_attn_layer_norm.bias")

            _ffn_name = _name + "ffn_prepost_wrapper/ffn/"
            new_var_dict[_ffn_name + "dense1/kernel"] = get_pyvar(_pyname + "fc1.weight").transpose()
            new_var_dict[_ffn_name + "dense1/bias"] = get_pyvar(_pyname + "fc1.bias")
            new_var_dict[_ffn_name + "dense2/kernel"] = get_pyvar(_pyname + "fc2.weight").transpose()
            new_var_dict[_ffn_name + "dense2/bias"] = get_pyvar(_pyname + "fc2.bias")
            new_var_dict[_name + "ffn_prepost_wrapper/ln/gamma"] = get_pyvar(_pyname + "final_layer_norm.weight")
            new_var_dict[_name + "ffn_prepost_wrapper/ln/beta"] = get_pyvar(_pyname + "final_layer_norm.bias")

        if not cfgs["decoder.post_normalize"]:
            new_var_dict[f"{tf_prefix}/TransformerDecoder/output_ln/gamma"] = get_pyvar("decoder.layer_norm.weight")
            new_var_dict[f"{tf_prefix}/TransformerDecoder/output_ln/beta"] = get_pyvar("decoder.layer_norm.bias")

        ckpt_saver = tf.train.Checkpoint(**{
            name: tf.Variable(initial_value=numpy_var, trainable=False,
                              name=name, dtype="float32")
            for name, numpy_var in new_var_dict.items()})
        save_ckpt = os.path.join(save_path, "ckpt")
        logging.info(f"Saving checkpoint to {save_ckpt}")
        ckpt_saver.save(save_ckpt)



import importlib
import os

from neurst.training.callbacks import (CentralizedCallback, CustomCheckpointCallback, LearningRateScheduler,
                                       MetricReductionCallback)
from neurst.training.validator import Validator
from neurst.utils.registry import setup_registry

build_validator, register_validator = setup_registry(Validator.REGISTRY_NAME, base_class=Validator,
                                                     verbose_creation=True)

__all__ = [
    "CentralizedCallback",
    "CustomCheckpointCallback",
    "LearningRateScheduler",
    "MetricReductionCallback",

    "Validator",
    "register_validator",
    "build_validator"
]

models_dir = os.path.dirname(__file__)
for file in os.listdir(models_dir):
    path = os.path.join(models_dir, file)
    if not file.startswith('_') and not file.startswith('.') and file.endswith('.py'):
        model_name = file[:file.find('.py')] if file.endswith('.py') else file
        module = importlib.import_module('neurst.training.' + model_name)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import sys

from neurst.utils.flags_core import ModuleFlag
from neurst.utils.hparams_sets import get_hyper_parameters
from neurst.utils.registry import REGISTRIES


def cli_main():
    if len(sys.argv) == 1 or (len(sys.argv) == 2 and (sys.argv[1] in ["help", "--help", "-h"])):
        print("Usage: ")
        print("    >> python3 -m neurst.cli.view_registry registry_name")
        print("           Show registered classes and their aliases.")
        print()
        print("    >> python3 -m neurst.cli.view_registry registry_name class_name")
        print("           Show detailed parameters of the class.")
        print()
        print("All registry names: ")
        for k in REGISTRIES:
            print(f"    - {k}")
        exit()

    registry_name = sys.argv[1].lower()
    if registry_name not in REGISTRIES:
        print(f"Unknown registry name: {registry_name}")
    elif len(sys.argv) == 2:
        print(f"All registered {registry_name}(s): ")
        clsname2alias = {}
        for name, cls in REGISTRIES[registry_name].items():
            clsname = cls.__name__
            if clsname not in clsname2alias:
                clsname2alias[clsname] = []
            clsname2alias[clsname].append(name)
        if registry_name == "hparams_set":
            for k in clsname2alias:
                print(f"    - {k}")
        else:
            print("    |  Class  |  Aliases  |")
            for k, v in clsname2alias.items():
                print("    |  {}  |  {}  |".format(k, ", ".join(v)))
    elif len(sys.argv) == 3:
        detail_name = sys.argv[2]
        if registry_name == "hparams_set":
            hparams = get_hyper_parameters(detail_name)
            if len(hparams) == 0:
                print(f"Unknown hparams_set: {detail_name}")
            else:
                print(f"Pre-defined hyperparameters set of `{detail_name}`: ")
                print(json.dumps(get_hyper_parameters(detail_name), indent=4))
        elif detail_name not in REGISTRIES[registry_name]:
            print(f"Unknown class: {detail_name} under `{registry_name}`")
        else:
            if hasattr(REGISTRIES[registry_name][detail_name], "class_or_method_args"):
                flags = []
                module_flags = []
                for f in REGISTRIES[registry_name][detail_name].class_or_method_args():
                    if isinstance(f, ModuleFlag):
                        module_flags.append(f)
                    else:
                        flags.append(f)
                if len(flags) > 0:
                    print(f"Flags for {detail_name}:")
                    print("    |  flag  |  type  |  default  |  help  |")
                    for f in flags:
                        print(f"    |  {f.name}  |  {str(f.dtype)}  |  {f.default}  |  {f.help}  |")
                if len(module_flags) > 0:
                    print(f"Dependent modules for {detail_name}: ")
                    print("    |  name  |  module  |  help  |")
                    for f in module_flags:
                        print(f"    |  {f.name}  |  {f.module_name}  |  {f.help}  |")

            else:
                print(f"No flags defined for `{detail_name}` ({registry_name})")


if __name__ == "__main__":
    cli_main()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Implements transformer encoders as described in https://arxiv.org/abs/1706.03762. """
import tensorflow as tf

from neurst.layers import build_transformer_component, layer_utils
from neurst.layers.attentions.multi_head_attention import MultiHeadSelfAttention
from neurst.layers.common_layers import TransformerFFN
from neurst.layers.encoders import register_encoder
from neurst.layers.encoders.encoder import Encoder


@register_encoder
class TransformerEncoder(Encoder):
    """ Defines transformer encoders as described
    in https://arxiv.org/abs/1706.03762. """

    def __init__(self,
                 num_layers,
                 hidden_size,
                 num_attention_heads,
                 filter_size,
                 ffn_activation="relu",
                 attention_dropout_rate=0.,
                 attention_type="dot_product",
                 ffn_dropout_rate=0.,
                 layer_postprocess_dropout_rate=0.,
                 layer_postprocess_epsilon=1e-6,
                 post_normalize=False,
                 return_all_layers=False,
                 name=None):
        """ Initializes the transformer encoders.

        Args:
            num_layers: The number of stacked layers.
            hidden_size: The number of hidden units.
            num_attention_heads: The number of self attention heads.
            filter_size: The filter size of ffn layer.
            ffn_activation: The activation function of ffn layer.
            ffn_dropout_rate: The dropout rate for ffn layer.
            attention_dropout_rate: The dropout rate for ffn layer.
            attention_type: The self attention type.
            layer_postprocess_dropout_rate: The dropout rate for each layer post process.
            layer_postprocess_epsilon: The epsilon for layer norm.
            post_normalize: Whether to apply layernorm after each block.
            return_all_layers: Whether to return all encoding layers.
            name: The name of this encoder.
        """
        super(TransformerEncoder, self).__init__(
            num_layers=num_layers, hidden_size=hidden_size,
            num_attention_heads=num_attention_heads,
            filter_size=filter_size, ffn_activation=ffn_activation,
            ffn_dropout_rate=ffn_dropout_rate,
            attention_type=attention_type,
            attention_dropout_rate=attention_dropout_rate,
            layer_postprocess_dropout_rate=layer_postprocess_dropout_rate,
            layer_postprocess_epsilon=layer_postprocess_epsilon,
            post_normalize=post_normalize,
            name=name or self.__class__.__name__)
        self._stacking_layers = []
        assert post_normalize or (not post_normalize and not return_all_layers), (
            "`return_all_layers` is only available when `post_normalize`=True.")
        self._return_all_layers = return_all_layers

    def build(self, input_shape):
        """ Builds the transformer encoder layer. """
        params = self.get_config()
        for _ in range(params["num_layers"]):
            self._stacking_layers.append([
                build_transformer_component({
                    "base_layer.class": MultiHeadSelfAttention.__name__,
                    "base_layer.params": dict(
                        num_heads=params["num_attention_heads"],
                        num_units=params["hidden_size"],
                        attention_dropout_rate=params["attention_dropout_rate"],
                        attention_type=params["attention_type"],
                        name="self_attention"
                    )},
                    dropout_rate=params["layer_postprocess_dropout_rate"],
                    epsilon=params["layer_postprocess_epsilon"],
                    pre_norm=(not params["post_normalize"])),
                build_transformer_component({
                    "base_layer.class": TransformerFFN.__name__,
                    "base_layer.params": dict(
                        filter_size=params["filter_size"],
                        output_size=params["hidden_size"],
                        dropout_rate=params["ffn_dropout_rate"],
                        activation=params["ffn_activation"],
                        name="ffn")},
                    dropout_rate=params["layer_postprocess_dropout_rate"],
                    epsilon=params["layer_postprocess_epsilon"],
                    pre_norm=(not params["post_normalize"]))
            ])
        if not params["post_normalize"]:
            self._output_norm_layer = tf.keras.layers.LayerNormalization(
                epsilon=params["layer_postprocess_epsilon"],
                dtype="float32", name="output_ln")
        super(TransformerEncoder, self).build(input_shape)

    def call(self, inputs, inputs_padding, is_training=True):
        """ Encodes the inputs.

        Args:
            inputs: The embedded input, a float tensor with shape
                [batch_size, max_length, embedding_dim].
            inputs_padding: A float tensor with shape [batch_size, max_length],
                indicating the padding positions, where 1.0 for padding and
                0.0 for non-padding.
            is_training: A bool, whether in training mode or not.

        Returns:
            The encoded output with shape [batch_size, max_length, hidden_size]
        """
        # [batch_size, max_length], FLOAT_MIN for padding, 0.0 for non-padding
        all_layers = []
        self_attention_bias = layer_utils.input_padding_to_bias(inputs_padding)
        x = inputs
        if is_training:
            x = tf.nn.dropout(x, rate=self.get_config()[
                "layer_postprocess_dropout_rate"])
        for idx, layer in enumerate(self._stacking_layers):
            self_attention_layer = layer[0]
            ffn_layer = layer[1]
            with tf.name_scope("layer_{}".format(idx)):
                # self attention layer
                x = self_attention_layer(
                    x,  # x as query
                    bias=self_attention_bias,
                    is_training=is_training)
                # ffn
                x = ffn_layer(x, is_training=is_training)
                all_layers.append(x)
        if self.get_config()["post_normalize"]:
            if self._return_all_layers:
                return all_layers
            return x
        return self._output_norm_layer(x)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math

import tensorflow as tf

from neurst.utils.activations import gelu
from neurst.utils.configurable import extract_constructor_params


class AudioConvSubsamplingLayer(tf.keras.layers.Layer):
    """ Subsampling for audio features. """

    def __init__(self,
                 embedding_dim,
                 channels=256,
                 kernel_size=3,
                 strides=2,
                 layer_norm=True,
                 name=None):
        """ Initializes the layer for subsample the audio feature.

        Args:
            embedding_dim: An int scalar, the embedding dimension.
            channels: The channel size of the convolution layer.
            kernel_size: The kernel size of the convolution layer.
            strides: The stride size of the convolution layer.
            layer_norm: Whether to apply layer normalization.
            verbose: A boolean, whether to logging the parameters.
            name: The name of the layer.
        """
        self._params = extract_constructor_params(locals(), verbose=True)
        super(AudioConvSubsamplingLayer, self).__init__(name=name)
        self._embedding_dim = embedding_dim
        self._channels = channels
        self._kernel_size = kernel_size
        self._layer_norm = layer_norm
        self._strides = strides

    @property
    def embedding_dim(self):
        return self._embedding_dim

    def get_config(self):
        return self._params

    def build(self, input_shape):
        """ Builds the layer."""
        self._conv_layer1 = tf.keras.layers.Conv2D(
            filters=self._channels,
            kernel_size=(self._kernel_size, self._kernel_size),
            strides=(self._strides, self._strides),
            padding="VALID",
            activation=None,
            name="conv1")
        self._conv_layer2 = tf.keras.layers.Conv2D(
            filters=self._channels,
            kernel_size=(self._kernel_size, self._kernel_size),
            strides=(self._strides, self._strides),
            padding="VALID",
            activation=None,
            name="conv2")
        if self._layer_norm:
            self._norm_layer1 = tf.keras.layers.LayerNormalization(
                epsilon=1e-6, dtype="float32", name="ln1")
            self._norm_layer2 = tf.keras.layers.LayerNormalization(
                epsilon=1e-6, dtype="float32", name="ln2")
        self._dense_layer = tf.keras.layers.Dense(
            self._embedding_dim,
            activation=None,
            use_bias=True,
            name="output_dense")
        super(AudioConvSubsamplingLayer, self).__init__(input_shape)

    def call(self, inputs, **kwargs):
        """ Gets token embeddings or computes logits.

        Args:
            inputs: An float tensor with shape [batch_size, length, feature_dim, channels].

        Returns:
            A float tensor with shape [batch, new_length, new_feature_dim].
        """
        _ = kwargs
        audio_feature_dim = inputs.get_shape().as_list()[2]
        assert inputs.get_shape().ndims == 4
        manual_padding = [[0, 0], [self._kernel_size // 2, self._kernel_size // 2],
                          [self._kernel_size // 2, self._kernel_size // 2], [0, 0]]
        conv1 = self._conv_layer1(tf.pad(inputs, manual_padding))
        if self._layer_norm:
            conv1 = self._norm_layer1(conv1)
        conv1 = tf.nn.relu(conv1)
        conv2 = self._conv_layer2(tf.pad(conv1, manual_padding))
        if self._layer_norm:
            conv2 = self._norm_layer2(conv2)
        conv2 = tf.nn.relu(conv2)
        new_feature_dim = ((audio_feature_dim + self._strides - 1)
                           // self._strides + self._strides - 1) // self._strides * self._channels
        conv2_reshape = tf.reshape(conv2, tf.concat([tf.shape(conv2)[:2], [new_feature_dim]], axis=0))
        output = self._dense_layer(conv2_reshape)
        return output


class Wav2vec2ConvBlock(tf.keras.layers.Layer):
    """ The convolution block of Wav2vec2FeatureExtractor. """

    def __init__(self, dim, kernel, stride, dropout_rate=0.,
                 use_bias=False, norm_type=None, name=None):
        """

        Args:
            dim: The output dimension of this convolution layer.
            kernel: The kernel size.
            stride: The stride.
            dropout_rate: The dropout rate.
            use_bias: Whether to include bias in conv encoder.
            norm_type: The type of layer normalization, "layer" or "group" or None.
            name: The name of this layer
        """
        self._params = extract_constructor_params(locals(), verbose=False)
        super(Wav2vec2ConvBlock, self).__init__(name=name)
        self._dim = dim
        self._kernel = kernel
        self._stride = stride
        self._dropout_rate = dropout_rate
        self._use_bias = use_bias
        self._norm_type = norm_type

    def get_config(self):
        return self._params

    def build(self, input_shape):
        """ Builds the layer."""
        self._conv_layer = tf.keras.layers.Conv1D(
            self._dim, kernel_size=self._kernel,
            strides=self._stride, use_bias=self._use_bias,
            kernel_initializer=tf.keras.initializers.HeNormal(),
            name="conv")
        self._norm_layer = None
        if self._norm_type == "layer":
            self._norm_layer = tf.keras.layers.LayerNormalization(
                epsilon=1.e-5, dtype="float32", name="ln")
        elif self._norm_type is not None:
            assert self._norm_type == "group"
            import tensorflow_addons as tfa
            self._norm_layer = tfa.layers.GroupNormalization(
                self._dim, epsilon=1.e-5, axis=-1, dtype="float32", name="gn")

    def call(self, inputs, is_training=False):
        """ Applies the convolution.

        Args:
            inputs: A Tensor of shape [batch, width, channels]
            is_training: Whether is under training.

        Returns:
            The convolution output.
        """
        conv_out = self._conv_layer(inputs)
        if is_training and self._dropout_rate > 0:
            conv_out = tf.nn.dropout(conv_out, rate=self._dropout_rate)
        if self._norm_layer is not None:
            conv_out = self._norm_layer(conv_out)
        # The original fairseq wav2vec use the non-approximated version
        return gelu(conv_out, non_approximate=True)


class Wav2vec2FeatureExtractor(tf.keras.layers.Layer):
    """ Subsampling for raw audios. """

    def __init__(self,
                 conv_layers,
                 dropout=0.0,
                 mode="default",
                 conv_bias=False,
                 verbose=False,
                 name=None):
        """ Initializes wav2vec2's convolution layers.

        Args:
            conv_layers: A list of convolution layers, each of which is in form of [dim, kernel, stride].
            dropout: The dropout rate of each conv layer.
            mode: The mode for feature extractor. "default" has a single group norm with d
                groups in the first conv block, whereas layer_norm has layer norms in
                every block (meant to use with normalize=True)
            conv_bias: Whether to include bias in conv encoder.
            verbose: A boolean, whether to logging the parameters.
            name: The name of the layer.
        """
        self._params = extract_constructor_params(locals(), verbose=True)
        super(Wav2vec2FeatureExtractor, self).__init__(name=name)
        self._conv_layers_setting = conv_layers
        self._dropout = dropout
        self._mode = mode
        self._conv_bias = conv_bias
        self._conv_layers = []

    def get_config(self):
        return self._params

    def build(self, input_shape):
        """ Builds the layer."""
        for i, (dim, kernel, stride) in enumerate(self._conv_layers_setting):
            norm_type = None
            if self._mode == "layer_norm":
                norm_type = "layer"
            elif self._mode == "default" and i == 0:
                norm_type = "group"
            self._conv_layers.append(Wav2vec2ConvBlock(
                dim, kernel, stride, dropout_rate=self._dropout,
                use_bias=self._conv_bias, norm_type=norm_type, name=f"conv_block{i}"))
        super(Wav2vec2FeatureExtractor, self).__init__(input_shape)

    def call(self, inputs, is_training=False):
        """ Applies the convolutional feature extration.

        Args:
            inputs: An float tensor with shape [batch_size, width].
            is_training: Whether is under training.

        Returns:
            A float tensor with shape [batch, new_width, output_channels].
        """
        output = tf.expand_dims(inputs, axis=-1)
        for conv_layer in self._conv_layers:
            output = conv_layer(output, is_training=is_training)
        return output


class PositionalConv(tf.keras.layers.Layer):
    """ Defines the pos_conv described in Wav2vec2. """

    def __init__(self, dim, kernel_size, groups,
                 dropout=0, name=None):
        super(PositionalConv, self).__init__(name=name)
        self._dim = dim
        self._kernel_size = kernel_size
        self._groups = groups
        self._dropout = dropout

    def build(self, input_shape):
        class _Wav2vec2ConvWeightNorm(tf.keras.layers.Wrapper):
            """Performs weight normalization.
            This is modified from tfa.layers.WeightNormalization.

            """

            def __init__(self, layer: tf.keras.layers, **kwargs):
                super(_Wav2vec2ConvWeightNorm, self).__init__(layer, **kwargs)
                self._track_trackable(layer, name="layer")

            def build(self, input_shape):
                """Build `Layer`"""
                input_shape = tf.TensorShape(input_shape)
                self.input_spec = tf.keras.layers.InputSpec(shape=[None] + input_shape[1:])

                if not self.layer.built:
                    self.layer.build(input_shape)

                kernel_layer = self.layer
                kernel = kernel_layer.kernel

                self.layer_depth = int(kernel.shape[0])
                self.kernel_norm_axes = list(range(kernel.shape.rank - 1))
                self.g = self.add_weight(
                    name="g",
                    shape=(self.layer_depth,),
                    initializer="ones",
                    # dtype=kernel.dtype,
                    trainable=True,
                )
                self.v = kernel
                self._initialized = self.add_weight(
                    name="initialized",
                    shape=None,
                    initializer="zeros",
                    dtype=tf.dtypes.bool,
                    trainable=False)
                self.built = True

            def call(self, inputs):
                """Call `Layer`"""

                def _do_nothing():
                    return tf.identity(self.g)

                def _update_weights():
                    # Ensure we read `self.g` after _update_weights.
                    with tf.control_dependencies(self._initialize_weights(inputs)):
                        return tf.identity(self.g)

                g = tf.cond(self._initialized, _do_nothing, _update_weights)

                with tf.name_scope("compute_weights"):
                    # Replace kernel by normalized weight variable.
                    kernel = tf.transpose(tf.nn.l2_normalize(tf.transpose(self.v, [2, 1, 0]),
                                                             axis=self.kernel_norm_axes) * g, [2, 1, 0])
                    self.layer.kernel = kernel
                    update_kernel = tf.identity(self.layer.kernel)

                    # Ensure we calculate result after updating kernel.
                    with tf.control_dependencies([update_kernel]):
                        outputs = self.layer(inputs)
                        return outputs

            def compute_output_shape(self, input_shape):
                return tf.TensorShape(self.layer.compute_output_shape(input_shape).as_list())

            def _initialize_weights(self, inputs):
                """Initialize weight g.

                The initial value of g could either from the initial value in v,
                or by the input value if self.data_init is True.
                """
                with tf.control_dependencies(
                    [
                        tf.debugging.assert_equal(  # pylint: disable=bad-continuation
                            self._initialized, False, message="The layer has been initialized."
                        )
                    ]
                ):
                    assign_tensors = self._init_norm()
                    assign_tensors.append(self._initialized.assign(True))
                    return assign_tensors

            def _init_norm(self):
                """Set the weight g with the norm of the weight vector."""
                with tf.name_scope("init_norm"):
                    v_flat = tf.reshape(self.v, [-1, self.layer_depth])
                    v_norm = tf.linalg.norm(v_flat, axis=0)
                    g_tensor = self.g.assign(tf.cast(tf.reshape(v_norm, (self.layer_depth,)), "float32"))
                    return [g_tensor]

        self._pos_conv = _Wav2vec2ConvWeightNorm(
            tf.keras.layers.Conv1D(
                self._dim, kernel_size=self._kernel_size, groups=self._groups,
                kernel_initializer=tf.initializers.random_normal(
                    mean=0, stddev=math.sqrt(4. / (self._kernel_size * self._dim))),
                name="conv"), name="wn")
        self._norm_layer = tf.keras.layers.LayerNormalization(
            epsilon=1.e-5, dtype="float32", name="ln")

    def call(self, inputs, inputs_padding=None, is_training=False):
        """

        Args:
            inputs: A tensor with shape [batch, time, channels]
            inputs_padding: A tensor with the same shape as `inputs`, where 1. denotes the padding.

        Returns:

        """
        if inputs_padding is not None:
            inputs = inputs * (1. - tf.expand_dims(tf.cast(inputs_padding, dtype=inputs.dtype), -1))
        padding_num = self._kernel_size // 2
        x = tf.pad(inputs, [[0, 0], [padding_num, padding_num], [0, 0]])
        x_conv = self._pos_conv(x)[:, : tf.shape(inputs)[1] - self._kernel_size % 2, :]
        x_conv = gelu(x_conv, non_approximate=True) + inputs
        x_conv = self._norm_layer(x_conv)
        if is_training:
            x_conv = tf.nn.dropout(x_conv, self._dropout)
        return x_conv



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

import tensorflow as tf
from absl import logging

from neurst.criterions import Criterion, build_criterion
from neurst.data.datasets.multiple_dataset import MultipleDataset
from neurst.exps import register_exp
from neurst.exps.base_experiment import BaseExperiment
from neurst.models.model_utils import summary_model_variables
from neurst.training import training_utils
from neurst.utils import compat
from neurst.utils.checkpoints import restore_checkpoint_if_possible
from neurst.utils.flags_core import ModuleFlag


@register_exp(["eval"])
class Evaluator(BaseExperiment):
    """ For evaluating on criterion. """

    def __init__(self, args, **kwargs):
        """ Initializes a util class for evaluating neural models. """
        super(Evaluator, self).__init__(**kwargs)
        self._criterion = build_criterion(args)

    @staticmethod
    def class_or_method_args():
        return [
            ModuleFlag(Criterion.REGISTRY_NAME, help="The criterion for evaluation."),
        ]

    @staticmethod
    def build_evaluation_model(task, model, criterion):
        """ Build keras model for evaluation.

        Args:
            task: The task object.
            model: An instance of neurst.models.model.BaseModel.
            criterion: The criterion for evaluation.

        Returns: the evaluation model.
        """
        inps = task.create_inputs(compat.ModeKeys.EVAL)
        formatted_inps = task.example_to_input(inps, compat.ModeKeys.EVAL)
        model_out = model(formatted_inps, is_training=False)
        keras_model = tf.keras.Model(inps, criterion(formatted_inps, model_out))
        return keras_model

    def run(self):
        """ Evaluation on a existing model.

        Step 1: Build model.
        Step 2: Builds evaluation dataset.
        Step 3: Restore checkpoints.
        Step 4: Evaluate and reduce metric.
        """

        with training_utils.get_strategy_scope(self.strategy):
            tfds = training_utils.build_datasets(compat.ModeKeys.EVAL, self.strategy,
                                                 self.custom_dataset, self.task)
            keras_model = self.build_evaluation_model(self.task, self.model, self._criterion)
            keras_model.summary()
            summary_model_variables(keras_model)
            # Step 4: Restore checkpoints.
            stat = restore_checkpoint_if_possible(self.model, self.model_dir)
            if not stat:
                logging.info(f"WARNING: Fail to restore checkpoint from {self.model_dir}. "
                             "We assume this was done on purpose. ")
        # Step 5: Evaluate and reduce metric.
        start_time = time.time()
        results, avg_res, whole_res = training_utils.reduce_eval_results(
            self._criterion, self.custom_dataset, training_utils.make_predictions(
                self.strategy, keras_model, tfds, self.custom_dataset))
        logging.info("Evaluation elapsed: %.2fs", time.time() - start_time)

        def _display(res, name=None):
            if name:
                logging.info(f"Evaluation Results ({name}):")
            for k, v in res.items():
                logging.info("   %s: %.2f", k, v)

        if not isinstance(self.custom_dataset, MultipleDataset):
            _display(results)
        else:
            for name, res in results.items():
                _display(res, name)
            _display(avg_res, f"on average by weights {self.custom_dataset.sample_weights}")
            _display(whole_res, "mixed")



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from neurst.data.text import register_tokenizer
from neurst.data.text.tokenizer import Tokenizer


@register_tokenizer("moses")
class MosesTokenizer(Tokenizer):

    def __init__(self, language, glossaries=None,
                 aggressive_dash_splits=True, escape=False):
        super(MosesTokenizer, self).__init__(
            language=language, glossaries=glossaries)
        self._aggressive_dash_splits = aggressive_dash_splits
        self._escape = escape
        try:
            from sacremoses import MosesDetokenizer as MDetok
            from sacremoses import MosesTokenizer as MTok
            self._tok = MTok(lang=self.language)
            self._detok = MDetok(lang=self.language)
        except ImportError:
            raise ImportError('Please install Moses tokenizer with: pip3 install sacremoses')

    def tokenize(self, text, return_str=False):
        return self._tok.tokenize(self._convert_to_str(text),
                                  aggressive_dash_splits=self._aggressive_dash_splits,
                                  return_str=return_str,
                                  escape=self._escape,
                                  protected_patterns=self._glossaries)

    def detokenize(self, text, return_str=True):
        return self._detok.detokenize(self._convert_to_list(text),
                                      return_str=return_str,
                                      unescape=True)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod

import six

from neurst.data.datasets.dataset import Dataset


@six.add_metaclass(ABCMeta)
class TextGenDataset(Dataset):
    """ The abstract dataset for text generation, which must implement `get_targets` function. """

    def __init__(self):
        self._targets = None
        super(TextGenDataset, self).__init__()

    @property
    @abstractmethod
    def status(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def build_iterator(self, map_func=None, shard_id=0, total_shards=1):
        """ Returns the iterator of the dataset.

        Args:
            map_func: A function mapping a dataset element to another dataset element.
            shard_id: Generator yields on the `shard_id`-th shard of the whole dataset.
            total_shards: The number of total shards.
        """
        raise NotImplementedError

    @property
    def targets(self):
        """ Returns a list of target texts. """
        return self._targets



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl import logging

from neurst.models.model import BaseModel


class EncoderDecoderEnsembleModel(BaseModel):
    """ Defines the basic sequence to sequence model. """

    def __init__(self, encoder_decoder_models):
        """ Initializes the ensemble model.

        Args:
            encoder_decoder_models: A list of encoder decoder models.
        """
        logging.info("EncoderDecoderEnsembleModel is only available for generation.")
        super(EncoderDecoderEnsembleModel, self).__init__(None, None)
        self._encoder_decoder_models = encoder_decoder_models

    @property
    def model_num(self):
        return len(self._encoder_decoder_models)

    @classmethod
    def new(cls, encoder_decoder_models):
        """ Creates keras model and model object.

        Args:
            encoder_decoder_models: A list of encoder decoder models.

        Returns:
            An encoder decoder ensemble model.
        """
        logging.info("Create model: {}".format(cls.__name__))
        return cls(encoder_decoder_models)

    def get_symbols_to_logits_fn(self, inputs, is_training=False, is_inference=True,
                                 decode_padded_length=None):
        """ Prepares for decoding.

        Args:
            inputs: A dict of model inputs.
                - src: int tensor with shape [batch_size, src_input_length].
                - src_padding: float tensor with shape [batch_size, src_input_length].
                - trg_input: int tensor with shape [batch_size, trg_length].
            is_training: A bool, whether in training mode or not.
            is_inference: A bool, whether in generation mode or not.
            decode_padded_length: The maximum decoding length when inference, for creating
                static-shape cache.

        Returns:  A tuple of (decoding_internal_states, decoder_input, symbol_to_logit_fn)
        """
        _ = is_training
        _ = is_inference
        is_training = False
        is_inference = True

        symbols_to_logits_fn_per_model = []
        final_generation_initializer = None
        for model in self._encoder_decoder_models:
            model_symbols_to_logits_fn, model_generation_initializer = model.get_symbols_to_logits_fn(
                inputs, is_training, is_inference, decode_padded_length=decode_padded_length)
            symbols_to_logits_fn_per_model.append(model_symbols_to_logits_fn)
            if final_generation_initializer is None:
                final_generation_initializer = model_generation_initializer
                final_generation_initializer["decoder_internal_cache"] = [
                    model_generation_initializer["decoder_internal_cache"]]
            else:
                final_generation_initializer["decoder_internal_cache"].append(
                    model_generation_initializer["decoder_internal_cache"])

        def symbols_to_logits_fn(symbols, cache_per_model, time=None):
            """ Generate logits for next potential IDs

            Args:
                symbols: Current decoded sequence.
                cache_per_model: A list of dictionaries of values storing the encoder output, encoder-decoder
                    attention bias, and previous decoder attention values.
                time: Loop index, or None for transformer training

            Returns: The logits Tensor.
            """
            logits_per_model = [
                fn(symbols, cache, time=time) for fn, cache in
                zip(symbols_to_logits_fn_per_model, cache_per_model)]
            return logits_per_model

        return symbols_to_logits_fn, final_generation_initializer

    def call(self, inputs, is_training=True):
        raise NotImplementedError("No need to implement call function for EncoderDecoderEnsembleModel.")



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl import logging

from neurst.data.data_pipelines import register_data_pipeline
from neurst.data.data_pipelines.data_pipeline import DataPipeline
from neurst.data.text import build_tokenizer
from neurst.data.text.vocab import PaddingMode, Vocab


@register_data_pipeline("simple_text")
class TextDataPipeline(DataPipeline, Vocab):

    def __init__(self,
                 vocab_path,
                 language="en",
                 tokenizer=None,
                 subtokenizer=None,
                 subtokenizer_codes=None,
                 glossaries=None,
                 reverse_sequence=False,
                 **kwargs):
        """ Initializes the data pipeline for text data.

        Args:
            language: The language.
            vocab_path: The path to the vocabulary file.
            tokenizer: The tokenizer name.
            subtokenizer: The name of tokenizer for subword encoding.
            subtokenizer_codes: The subword codes.
            glossaries: The glossaries that will not be split by tokenizer/subtokenizer.
            reverse_sequence: A bool, whether to reverse the sequence.
        """
        DataPipeline.__init__(self, vocab_path=vocab_path, language=language,
                              tokenizer=tokenizer, subtokenizer=subtokenizer,
                              subtokenizer_codes=subtokenizer_codes,
                              glossaries=glossaries,
                              reverse_sequence=reverse_sequence,
                              **kwargs)
        self._language = language
        self._reverse_sequence = reverse_sequence
        self._tokenizer = build_tokenizer(tokenizer, language=language, glossaries=glossaries)
        self._subtokenizer = None
        self._subtokenizer = build_tokenizer(
            subtokenizer, language=language, glossaries=glossaries, vocabulary=vocab_path)
        if self._subtokenizer is not None:
            if subtokenizer_codes is None:
                logging.info("No codes provided for subtokenizer: {}. "
                             "We assume this was done on purpose.".format(subtokenizer))
            else:
                self._subtokenizer.init_subtokenizer(subtokenizer_codes)
        tokens = Vocab.load_tokens(vocab_path)
        unk_token = Vocab.get_unique(tokens, "<UNK>")
        bos_token = Vocab.get_unique(tokens, "<SEQ_BEG>")
        eos_token = Vocab.get_unique(tokens, "<SEQ_END>")
        assert unk_token != bos_token != eos_token
        Vocab.__init__(self, tokens, [unk_token, bos_token, eos_token], lowercase=False)
        self._eos_id = Vocab.map_token_to_id(self, eos_token)
        self._bos_id = Vocab.map_token_to_id(self, bos_token)
        self._unk_id = Vocab.map_token_to_id(self, unk_token)

    @property
    def meta(self):
        return {
            "language": self._language,
            "vocab_size": self.vocab_size,
            "eos_id": self._eos_id,
            "bos_id": self._bos_id,
            "unk_id": self._unk_id,
            "pad_id": self._eos_id,
            "padding_mode": PaddingMode.EOS_AS_PADDING
        }

    def process(self, input, is_processed=False):
        """ Process one data sample.

        Args:
            input: A text string.
            is_processed: Whether the data sample is already processed.

        Returns:
            A list of generated token IDs.
        """
        if not is_processed:
            if self._tokenizer:
                input = self._tokenizer.tokenize(input)
            if self._subtokenizer:
                input = self._subtokenizer.tokenize(input, return_str=False)
        if isinstance(input, str):
            input = input.split()
        token_ids = Vocab.map_token_to_id(self, input, unknown_default=self._unk_id)
        if self._reverse_sequence:
            token_ids = token_ids[::-1]
        return token_ids + [self._eos_id]

    def recover(self, input):
        """ Recover one data sample.

        Args:
            input: A list of token ids, the output of neural model.

        Returns:
            A string, the recovered text.
        """
        input = [int(x) for x in input]
        if input[0] == self._bos_id:
            input = input[1:]
        try:
            eos_pos = input.index(self._eos_id)
            input = input[:eos_pos]
        except ValueError:
            pass
        token_list = Vocab.map_id_to_token(self, input)
        if self._reverse_sequence:
            token_list = token_list[::-1]
        if self._subtokenizer is None:
            output = " ".join(token_list)
        else:
            output = self._subtokenizer.detokenize(token_list, return_str=True)
        if self._tokenizer:
            output = self._tokenizer.detokenize(output, return_str=True)
        return output



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy

from neurst.data.audio import FeatureExtractor, register_feature_extractor


@register_feature_extractor
class FloatIdentity(FeatureExtractor):

    def __init__(self, args):
        _ = args

    def __call__(self, signal, rate):
        if type(signal[0]) is float:
            return numpy.array(signal)
        return numpy.array(signal) / 32768.



import importlib

importlib.import_module("neurst.data.audio")
importlib.import_module("neurst.data.data_pipelines")
importlib.import_module("neurst.data.datasets")
importlib.import_module("neurst.data.text")



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod
from typing import Tuple

import six
import tensorflow as tf

from neurst.data.datasets import Dataset
from neurst.utils.configurable import copy_dict_list
from neurst.utils.flags_core import COMMON_DATA_ARGS


@six.add_metaclass(ABCMeta)
class Task(object):
    """ The task object binds the data pipelines. """

    REGISTRY_NAME = "task"

    def __init__(self, args):
        """ Initializes with configuration. """
        self._args = args

    def model_configs(self, model):
        return {
            "model.class": model.__class__.__name__,
            "model.params": copy_dict_list(model.args),
            "task.class": self.__class__.__name__,
            "task.params": self.get_config()
        }

    def create_inputs(self, mode):
        """ Creates keras input placeholders. """
        dtypes, signatures = self.inputs_signature(mode)
        inps = dict()
        for name in dtypes:
            inps[name] = tf.keras.layers.Input(tuple(signatures[name][1:]),
                                               dtype=dtypes[name], name=name)
        return inps

    @abstractmethod
    def inputs_signature(self, mode) -> Tuple[dict, dict]:
        """ Defines the types and signatures of the dataset inputs. """
        raise NotImplementedError("Task must implement the `inputs_signature` method.")

    @abstractmethod
    def example_to_input(self, batch_of_data: dict, mode) -> dict:
        """ Converts a batch of data into the model readable structure. """
        raise NotImplementedError("Task must implement the `example_to_input` method.")

    @abstractmethod
    def get_data_preprocess_fn(self, mode, data_status, args=None) -> callable:
        """ Returns a callable function that preprocess the data sample
            according to this task. """
        raise NotImplementedError("Task must implement the `get_data_preprocess_fn` method.")

    def get_data_postprocess_fn(self, mode) -> callable:
        """ Returns a callable function that postprocess the data sample
            according to this task. """
        raise NotImplementedError

    @abstractmethod
    def create_and_batch_tfds(self,
                              ds: Dataset,
                              mode,
                              args=None,
                              num_replicas_in_sync=1) -> tf.data.Dataset:
        """ Batch dataset. """
        raise NotImplementedError("Task must implement the `create_and_batch_tfds` method.")

    @abstractmethod
    def get_config(self):
        return {}

    @staticmethod
    def class_or_method_args():
        """ Returns a list of args for flag definition. """
        return [x for x in COMMON_DATA_ARGS]

    @abstractmethod
    def build_model(self, args, name=None):
        """Build a new model instance."""
        raise NotImplementedError("Task must implement the `build_model` method.")

    def build_metric_layer(self):
        """ Builds a list of metric layers for logging and tensorboard. """
        return []

    def get_eval_metric(self, args, name="metric"):
        """ Returns a neurst.metrics.metric.Metric object for evaluation."""
        return None



import importlib
import os

import tensorflow as tf

from neurst.utils.registry import setup_registry

LR_SCHEDULE_REGISTRY_NAME = "lr_schedule"
build_lr_schedule, register_lr_schedule = setup_registry(
    LR_SCHEDULE_REGISTRY_NAME, base_class=tf.keras.optimizers.schedules.LearningRateSchedule,
    verbose_creation=True)

models_dir = os.path.dirname(__file__)
for file in os.listdir(models_dir):
    path = os.path.join(models_dir, file)
    if not file.startswith('_') and not file.startswith('.') and file.endswith('.py'):
        model_name = file[:file.find('.py')] if file.endswith('.py') else file
        module = importlib.import_module('neurst.optimizers.schedules.' + model_name)



import numpy

from neurst.layers.modalities.text_modalities import WordEmbeddingSharedWeights


def test_word_embedding():
    embedding_layer = WordEmbeddingSharedWeights(
        embedding_dim=5, vocab_size=10,
        share_softmax_weights=False)
    inputs2d = numpy.random.randint(0, 9, size=(3, 4))
    inputs1d = numpy.random.randint(0, 9, size=(3,))
    emb_for_2d = embedding_layer(inputs2d)
    emb_for_1d = embedding_layer(inputs1d)
    assert len(embedding_layer.get_weights()) == 1
    assert numpy.sum(
        (embedding_layer.get_weights()[0][inputs1d] - emb_for_1d.numpy()) ** 2) < 1e-9
    assert numpy.sum(
        (embedding_layer.get_weights()[0][inputs2d] - emb_for_2d.numpy()) ** 2) < 1e-9
    assert "emb/weights" in embedding_layer.trainable_weights[0].name

    emb_shared_layer = WordEmbeddingSharedWeights(
        embedding_dim=5, vocab_size=10,
        share_softmax_weights=True)
    emb_for_2d = emb_shared_layer(inputs2d)
    emb_for_1d = emb_shared_layer(inputs1d)
    logits_for_2d = emb_shared_layer(emb_for_2d, mode="linear")
    logits_for_1d = emb_shared_layer(emb_for_1d, mode="linear")
    assert len(emb_shared_layer.get_weights()) == 2
    for w in emb_shared_layer.trainable_weights:
        if "bias" in w.name:
            bias = w
        else:
            weights = w
    assert numpy.sum(
        (weights.numpy()[inputs1d] - emb_for_1d.numpy()) ** 2) < 1e-9
    assert numpy.sum(
        (weights.numpy()[inputs2d] - emb_for_2d.numpy()) ** 2) < 1e-9
    assert numpy.sum(
        (numpy.dot(emb_for_2d.numpy(), numpy.transpose(weights.numpy())
                   ) + bias.numpy() - logits_for_2d.numpy()) ** 2) < 1e-9
    assert numpy.sum(
        (numpy.dot(emb_for_1d.numpy(), numpy.transpose(weights.numpy())
                   ) + bias.numpy() - logits_for_1d.numpy()) ** 2) < 1e-9
    assert "shared/weights" in weights.name
    assert "shared/bias" in bias.name


if __name__ == "__main__":
    test_word_embedding()



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl import logging

from neurst.layers import layer_utils
from neurst.layers.search import register_search_layer
from neurst.layers.search.sequence_search import SequenceSearch
from neurst.utils import compat
from neurst.utils.flags_core import Flag


def _calculate_length_penalty(lengths, alpha, dtype=None):
    """ Calculates length penalty, Referring
      to https://arxiv.org/abs/1609.08144.

    Args:
        lengths: The length tensor, with shape [n, ]
        alpha: The length penalty rate. Length penalty is given by
          (5+len(decode)/6) ^ -\alpha.
        dtype: The dtype for return.

    Returns: The length penalty tensor.
    """
    if dtype is None:
        dtype = compat.CUSTOM_GLOBAL_FLOATX
    if alpha is None or alpha < 0.0:
        return tf.constant(1., dtype=dtype) / tf.cast(lengths, dtype=dtype)
    return ((5.0 + tf.cast(lengths, dtype)) / 6.0) ** (-alpha)


class _StateKeys(object):
    """Keys to dictionary storing the state of the beam search loop."""

    # The decoding step.
    TIME_STEP = "TIME_STEP"

    # The decoder input ids with shape [batch_size * beam_size, ]
    INPUT_IDS = "INPUT_IDS"

    # Dictionary of cached values for each beam. The cache stores encoder
    # output (memory), encoder input paddings and the decoder attention
    # output from the previous iteration (usually for transformer)
    CACHE = "CACHE"

    # Flags indicating which sequences in the finished sequences are finished,
    # with shape [batch_size * beam_size, ]
    FINISHED_FLAGS = "FINISHED_FLAGS"

    # Log probabilities of each sequence with shape [batch_size * beam_size]
    LOG_PROBS = "LOG_PROBS"

    # The true decoding length of each beam
    DECODING_LENGTH = "LENGTH"

    # The predicted sequences, with shape [time_step, batch_size * beam_size]
    PREDICTED_IDS = "PREDICTED_IDS"


def _calculate_log_probs(state,
                         symbols_to_logits_fn,
                         eos_id,
                         unk_id,
                         ensemble_weights=None):
    """ Calculates one-step log probability.

    Finished beam will be masked and UNK will be masked
    if strategy == BASIC_NO_UNK.

    Args:
        state: A dictionary containing current state of beam search.
        symbols_to_logits_fn:
        eos_id: An int scalar, indicating the end-of-sentence token id, used to determine when a
            sequence has finished.
        unk_id: An int scalar, indicating the unknown token id.
        ensemble_weights: A list of float values, indicating the weights of each submodel's probability.

    Returns:
        A float tensor with the same shape as `logits`.
    """
    logits = symbols_to_logits_fn(
        state[_StateKeys.INPUT_IDS], state[_StateKeys.CACHE],
        state[_StateKeys.TIME_STEP])
    logits = tf.nest.flatten(logits)
    vocab_size = logits[0].get_shape().as_list()[-1]
    batch_beam_size = tf.shape(logits[0])[0]
    if len(logits) == 1:
        # [batch_size * beam_size, target_vocab_size]
        log_probs = tf.nn.log_softmax(logits[0])
    else:
        probs = tf.nest.map_structure(
            lambda x: tf.expand_dims(
                tf.reshape(tf.nn.softmax(x), shape=[-1]), axis=0),
            logits)
        original_shape = tf.shape(logits[0])
        # [num_models, xxx]
        probs = tf.concat(probs, axis=0)
        # [1, num_models]
        weights = tf.expand_dims(
            tf.convert_to_tensor(ensemble_weights, dtype=probs.dtype),
            axis=0)
        probs = tf.matmul(weights, probs)
        log_probs = tf.math.log(tf.reshape(probs, original_shape))

    # [batch_size * beam_size,]
    prev_finished_float = tf.cast(state[_StateKeys.FINISHED_FLAGS],
                                  log_probs.dtype)
    # mask the finished beam except only one entrance (target_eos_id)
    #   [target_vocab_size, ]: [float_min, float_min, float_min, ..., 0]
    #   this forces the beam with EOS continue to generate EOS
    finished_beam_bias = layer_utils.one_entry_bias(
        on_entry=eos_id, num_entries=vocab_size,
        on_value=0., off_value=compat.FLOAT_MIN,
        dtype=log_probs.dtype)
    # [batch_size * beam_size, target_vocab_size]: outer product
    finished_beam_bias = layer_utils.tile_tensor(
        finished_beam_bias, batch_beam_size, axis=0)
    finished_beam_bias *= tf.expand_dims(prev_finished_float, 1)
    # compute new probs, with finished flags & mask
    log_probs = log_probs * tf.expand_dims(1. - prev_finished_float, 1) + finished_beam_bias

    # we should use the trick for masking out the UNK in the probability
    if unk_id is not None:
        unk_beam_bias = layer_utils.one_entry_bias(
            on_entry=unk_id, num_entries=vocab_size,
            on_value=compat.FLOAT_MIN, off_value=0.,
            dtype=log_probs.dtype)
        unk_beam_bias = layer_utils.tile_tensor(
            unk_beam_bias, batch_beam_size, axis=0)
        log_probs += unk_beam_bias
    return log_probs


def _sample_next_word(state,
                      log_probs,
                      beam_size,
                      length_penalty):
    """ Generates next token ids.

    Args:
        state: A dictionary containing current state of beam search.
        log_probs: The log probs of current timestep with shape
            [batch_size * beam_size, vocab_size]
        beam_size: The beam width.
        length_penalty: A float scalar, defining the strength of
            length normalization.

    Returns: A tuple `(word_ids, beam_ids, next_log_probs, next_lengths)`, where
      `words_ids` is the ids of sampled word symbols; `beam_ids` indicates the index
      of beam which the symbol at the position is from; `next_log_probs` is the accumulated
      log probabilities of each beam; `next_lengths` is the decoding lengths of
      each beam.
      All of the Tensors have shape [batch_size * beam_size, ].
    """
    batch_beam_size = tf.shape(log_probs)[0]
    batch_size = batch_beam_size // beam_size
    vocab_size = log_probs.get_shape().as_list()[-1]
    # calculates the accumulated log probs
    #   [batch_size * beam_size, target_vocab_size]
    log_probs = log_probs + tf.expand_dims(state[_StateKeys.LOG_PROBS], 1)
    # new decoding length: [batch_size * beam_size, ]
    next_length = state[_StateKeys.DECODING_LENGTH] + 1 - tf.cast(
        state[_StateKeys.FINISHED_FLAGS], tf.int32)
    # calculates beam scores
    #  length_penalty: [batch_size * beam_size,]
    penalty_term = _calculate_length_penalty(
        next_length, length_penalty, dtype=log_probs.dtype)
    scores = log_probs * tf.expand_dims(penalty_term, axis=1)

    # flatten: [batch_size, beam_size * target_vocab_size]
    scores = tf.reshape(tf.reshape(scores, [-1]),
                        [batch_size, -1])
    scores_flat = tf.cond(
        state[_StateKeys.TIME_STEP] > 0,
        lambda: scores,  # time > 0: all
        lambda: tf.slice(scores, [0, 0],
                         [-1, vocab_size]))  # time = 0: first logits in each batch
    # [batch_size, beam_size] will restore top beam results
    sample_scores, sample_ids = tf.nn.top_k(scores_flat, k=beam_size)
    # flatten: [batch_size * beam_size,]
    sample_ids = tf.reshape(sample_ids, [-1])
    # because we do topk to scores with dim: [batch, beam * vocab]
    #   we need to cover the true word ids
    word_ids = tf.math.floormod(sample_ids, vocab_size)

    # find beam_ids, indicating the current position is from which beam
    #  with shape [batch_size, beam_size], e.g. [[0, 0, ...], [1, 1,...], ..., [batch_size,...] ]
    batch_pos = layer_utils.compute_batch_indices(batch_size, beam_size)
    #  beam_base_pos: [batch_size * beam_size,]: [0, 0, ..., beam, beam,..., 2beam, 2beam, ...]
    beam_base_pos = tf.reshape(batch_pos * beam_size, [-1])
    # compute new beam_ids, [batch_size * beam_size, ]
    beam_ids = tf.math.floordiv(sample_ids, vocab_size) + beam_base_pos

    # gather states according to beam_ids
    next_lengths = tf.gather(next_length, beam_ids)

    # we need to recover log_probs according to scores's topk ids
    #   [batch_size * beam_size * vocab_size, ]
    log_probs_flat = tf.reshape(log_probs, [-1])
    log_probs_index = beam_base_pos * vocab_size + sample_ids
    next_log_probs = tf.gather(log_probs_flat, log_probs_index)

    return word_ids, beam_ids, next_log_probs, next_lengths


def _extract_beam_results(state,
                          beam_size,
                          top_k,
                          length_penalty):
    """ Extract `top_k` hypothesis from beam search results.

    Args:
        state: A dictionary containing current state of beam search.
        beam_size: The beam width.
        top_k: The number of hypothesis of each sample with highest scores
            with be returned. It must be <= `beam_sze`.
        length_penalty: A float scalar, defining the strength of
            length normalization.

    Returns: A tuple of sorted hypothesis and corresponding scores.
        => sorted_hypothesis: [batch_size * top_k, max_len_of_hypothesis]
        => sorted_socres: [batch_size * top_k, ]
    """
    log_probs = state[_StateKeys.LOG_PROBS]
    penalty = _calculate_length_penalty(
        state[_StateKeys.DECODING_LENGTH],
        length_penalty, dtype=log_probs.dtype)
    scores = log_probs * penalty
    # [_batch * _beam, ] => [_batch, _beam]
    scores_shaped = tf.reshape(scores, [-1, beam_size])
    # [_batch, _top_k]
    top_scores, top_indices = tf.nn.top_k(scores_shaped, k=top_k)
    batch_beam_pos = layer_utils.compute_batch_indices(
        tf.shape(top_indices)[0], k=top_k) * beam_size
    # [_batch * _top_k, ]
    top_indices = tf.reshape(top_indices + batch_beam_pos, [-1])
    # [_batch * _top_k, timesteps]
    sorted_hypothesis = tf.gather(state[_StateKeys.PREDICTED_IDS], top_indices)
    sorted_scores = tf.reshape(top_scores, [-1])
    return sorted_hypothesis, sorted_scores


def sequence_beam_search(symbols_to_logits_fn,
                         generation_initializer,
                         top_k=1,
                         beam_size=4,
                         length_penalty=0.6,
                         extra_decode_length=50,
                         maximum_decode_length=256,
                         minimum_decode_length=0,
                         ensemble_weights=None,
                         padded_decode=False,
                         enable_unk=False,
                         dtype=None):
    """ Search for target subtokens with the largest probability.

    Args:
        symbols_to_logits_fn: A callable function that takes ids, cache and current
            timestep as arguments. The passed in arguments will have following shape:
                ids -> A tensor with shape [batch_size * beam_size,].
                cache -> A nested dictionary of tensors [batch_size * beawm_size, ...].
                index -> A scalar.
            The function must return a logits tensor with shape
                [batch_size * beam_size, vocab_size].
        generation_initializer: A dict from model, containing:
            -> decoder_input_ids: An int32 tensor with shape [batch_size,].
                    The initial input for the decoder.
            -> decoding_cache: A nested dictionary, containing attention memory and other
                    starting variables information.
            -> encoder_inputs_maxlen: An int32 scalar tensor, indicating the maximum length of encoding input
                    sequence.
            -> eos_id: An int scalar, indicating the end-of-sentence symbol id, used to determine when a
                    sequence has finished.
            -> unk_id: An int scalar, indicating the unknown token id.
        top_k: An int scalar, indicating the number of hypothesis each sample will generate.
        beam_size: An int scalar, indicating the number of beams.
        length_penalty: A float scalar, defining the strength of length normalization.
        extra_decode_length: An int scalar. The real search steps will be
            `encoder_inputs_maxlen` + `extra_decode_length`.
        maximum_decode_length: An int scalar, indicating the maximum decode length. Prediction
            outputs will be padded to this length.
        minimum_decode_length: An int scalar, indicating the minimum decode length.
        ensemble_weights: A list of float values, indicating the weights of each submodel's probability.
        padded_decode: Whether the autoregressive decoding runs with input data padded to the decode_max_length.
        enable_unk: Whether to enable the search method to generate UNK.
        dtype: A string or a tensorflow data type used for score computation. If None, auto set
            to the GLOBAL_FLOATX.

    Returns:
        A tuple of (hypothesis, scores):
            hypothesis: int32 tensor with shape [batch_size * top_k, decode_length]
            scores: float tensor with shape [batch_size * top_k]
    """
    decoder_input_ids = generation_initializer["decoder_input"]
    decoding_cache = generation_initializer["decoder_internal_cache"]
    encoder_inputs_maxlen = generation_initializer["encoder_inputs_maxlen"]
    eos_id = generation_initializer["eos_id"]
    unk_id = None if enable_unk else generation_initializer.get("unk_id", None)
    if dtype is None:
        dtype = tf.dtypes.as_dtype(compat.CUSTOM_GLOBAL_FLOATX)
    batch_size = tf.shape(decoder_input_ids)[0]
    initial_finished = tf.tile([False], [batch_size * beam_size])
    initial_decoder_input_ids = layer_utils.stack_beam_size(decoder_input_ids, beam_size)
    initial_time = tf.constant(0, dtype=tf.int32)
    initial_cache = tf.nest.map_structure(
        lambda x: layer_utils.stack_beam_size(x, beam_size), decoding_cache)

    initial_log_probs = tf.zeros_like(
        initial_decoder_input_ids, dtype=dtype)
    initial_length = tf.zeros_like(initial_decoder_input_ids, dtype=tf.int32)
    # [time, batch_size * beam_size]
    if padded_decode:
        initial_predicted_ids = tf.zeros([batch_size * beam_size, maximum_decode_length],
                                         dtype=tf.int32)
    else:
        initial_predicted_ids = tf.zeros([batch_size * beam_size, 0],
                                         dtype=tf.int32)
    # Create state dictionary
    search_state = {
        _StateKeys.TIME_STEP: initial_time,
        _StateKeys.INPUT_IDS: tf.cast(initial_decoder_input_ids, tf.int32),
        _StateKeys.CACHE: initial_cache,
        _StateKeys.FINISHED_FLAGS: initial_finished,
        _StateKeys.LOG_PROBS: initial_log_probs,
        _StateKeys.DECODING_LENGTH: initial_length,
        _StateKeys.PREDICTED_IDS: initial_predicted_ids
    }
    # Create state invariants for each value in the state dictionary. Each
    # dimension must be a constant or None. A None dimension means either:
    #   1) the dimension's value is a tensor that remains the same but may
    #      depend on the input sequence to the model (e.g. batch size).
    #   2) the dimension may have different values on different iterations.
    if padded_decode:
        search_state_shape_invariants = {
            _StateKeys.TIME_STEP: tf.TensorShape([]),
            _StateKeys.INPUT_IDS: tf.TensorShape([None]),
            _StateKeys.CACHE: tf.nest.map_structure(
                layer_utils.static_tensorshape, initial_cache),
            _StateKeys.FINISHED_FLAGS: tf.TensorShape([None]),
            _StateKeys.LOG_PROBS: tf.TensorShape([None]),
            _StateKeys.DECODING_LENGTH: tf.TensorShape([None]),
            _StateKeys.PREDICTED_IDS: tf.TensorShape([None, maximum_decode_length]),
        }
    else:
        search_state_shape_invariants = {
            _StateKeys.TIME_STEP: tf.TensorShape([]),
            _StateKeys.INPUT_IDS: tf.TensorShape([None]),
            _StateKeys.CACHE: tf.nest.map_structure(
                layer_utils.dynamic_tensorshape_except_last_dim, initial_cache),
            _StateKeys.FINISHED_FLAGS: tf.TensorShape([None]),
            _StateKeys.LOG_PROBS: tf.TensorShape([None]),
            _StateKeys.DECODING_LENGTH: tf.TensorShape([None]),
            _StateKeys.PREDICTED_IDS: tf.TensorShape([None, None]),
        }
    maximum_search_steps = tf.minimum(
        encoder_inputs_maxlen + extra_decode_length,
        maximum_decode_length)
    maximum_search_steps = tf.maximum(maximum_search_steps, minimum_decode_length)

    def search_step(state):
        """ Beam search step. """
        # [batch_size * beam_size, vocab_size]
        log_probs = _calculate_log_probs(
            state=state, symbols_to_logits_fn=symbols_to_logits_fn,
            eos_id=eos_id, unk_id=unk_id, ensemble_weights=ensemble_weights)
        # masking out the EOS in the probability when decoding length < min_length
        eos_beam_bias = layer_utils.one_entry_bias(
            on_entry=eos_id, num_entries=log_probs.get_shape().as_list()[-1],
            on_value=compat.FLOAT_MIN, off_value=0.,
            dtype=log_probs.dtype)
        eos_beam_bias = layer_utils.tile_tensor(eos_beam_bias, tf.shape(log_probs)[0], axis=0)
        log_probs = tf.cond(tf.less(state[_StateKeys.TIME_STEP], minimum_decode_length - 1),
                            lambda: log_probs + eos_beam_bias,
                            lambda: log_probs)

        # compute log probs and generate next token ids according to beam scores
        sample_ids, beam_ids, next_log_probs, next_length = _sample_next_word(
            state=state, log_probs=log_probs, beam_size=beam_size,
            length_penalty=length_penalty)
        # re-order beams by beam_ids
        next_predicted_ids = tf.gather(state[_StateKeys.PREDICTED_IDS], beam_ids)
        if padded_decode:
            next_predicted_ids = tf.transpose(tf.tensor_scatter_nd_update(
                tf.transpose(next_predicted_ids), [[state[_StateKeys.TIME_STEP]]],
                tf.expand_dims(sample_ids, axis=0)))
        else:
            next_predicted_ids = tf.concat([next_predicted_ids, tf.expand_dims(
                sample_ids, axis=1)], axis=1)
        next_cache = tf.nest.map_structure(
            lambda x: tf.gather(x, beam_ids), state[_StateKeys.CACHE])
        next_finished = tf.equal(eos_id, sample_ids)
        state.update({
            _StateKeys.TIME_STEP: state[_StateKeys.TIME_STEP] + 1,
            _StateKeys.INPUT_IDS: sample_ids,
            _StateKeys.CACHE: next_cache,
            _StateKeys.FINISHED_FLAGS: next_finished,
            _StateKeys.LOG_PROBS: next_log_probs,
            _StateKeys.DECODING_LENGTH: next_length,
            _StateKeys.PREDICTED_IDS: next_predicted_ids})
        return [state]

    # beam search step by step
    res = tf.while_loop(
        lambda state: tf.logical_and(
            tf.logical_not(tf.reduce_all(state[_StateKeys.FINISHED_FLAGS])),
            tf.less(state[_StateKeys.TIME_STEP], maximum_search_steps)),
        search_step,
        loop_vars=[search_state],
        shape_invariants=[search_state_shape_invariants],
        parallel_iterations=1)
    res = res[0]
    sorted_hypothesis, sorted_scores = _extract_beam_results(
        state=res, beam_size=beam_size, top_k=top_k,
        length_penalty=length_penalty)
    # padding to fixed length output
    if not padded_decode:
        sorted_hypothesis = tf.pad(
            sorted_hypothesis,
            paddings=tf.convert_to_tensor([
                [0, 0], [0, maximum_decode_length - res[_StateKeys.TIME_STEP]]],
                dtype=tf.int32),
            mode="CONSTANT",
            constant_values=eos_id)
    return sorted_hypothesis, sorted_scores


@register_search_layer
class BeamSearch(SequenceSearch):

    def __init__(self, args):
        super(BeamSearch, self).__init__()
        self._beam_size = args["beam_size"]
        self._length_penalty = args["length_penalty"]
        self._top_k = args["top_k"]
        if self._top_k > 1:
            logging.info("Initializes BeamSearch with top_k = {}. "
                         "Be careful if the resulted predictions are "
                         "passed to evaluation metrics. ".format(self._top_k))
        self._maximum_decode_length = args["maximum_decode_length"]
        self._extra_decode_length = args["extra_decode_length"]
        self._minimum_decode_length = args["minimum_decode_length"]
        if not self._minimum_decode_length:
            self._minimum_decode_length = 0
        self._ensemble_weights = args["ensemble_weights"]
        self._padded_decode = args["padded_decode"]
        self._enable_unk = args["enable_unk"]

    @staticmethod
    def class_or_method_args():
        return [
            Flag("beam_size", dtype=Flag.TYPE.INTEGER, default=4,
                 help="The beam width of beam search inference."),
            Flag("length_penalty", dtype=Flag.TYPE.FLOAT, default=0.6,
                 help="The length penalty of beam search inference."),
            Flag("top_k", dtype=Flag.TYPE.INTEGER, default=1,
                 help="The number of reserved predictions with top scores of beam search inference."),
            Flag("maximum_decode_length", dtype=Flag.TYPE.INTEGER, default=None,
                 help="The maximum decoding length of beam search inference."),
            Flag("minimum_decode_length", dtype=Flag.TYPE.INTEGER, default=0,
                 help="The minimum decoding length of beam search inference."),
            Flag("extra_decode_length", dtype=Flag.TYPE.INTEGER, default=50,
                 help="The extra decoding length versus source side for beam search inference. "
                      "The maximum decoding length of beam search inference will be "
                      "source_sequence_length + extra_decode_length if maximum_decode_length "
                      "if not provided."),
            Flag("padded_decode", dtype=Flag.TYPE.BOOLEAN, default=None,
                 help="Whether the autoregressive decoding runs with input data padded to "
                      "the decode_max_length. For TPU/XLA-GPU runs, this flag has to be "
                      "set due the static shape requirement. In addition, this method "
                      "will introduce unnecessary overheads which grow quadratically with "
                      "the max sequence length."),
            Flag("ensemble_weights", dtype=Flag.TYPE.STRING, default="average",
                 help="The weight scheme for model ensemble, which could be comma-separated numbers."),
            Flag("enable_unk", dtype=Flag.TYPE.BOOLEAN, default=None,
                 help="Whether to enable the search method to generating UNK."),
        ]

    def _get_ensemble_weights(self, ensemble_weights):
        """ Creates ensemble weights from `ensemble_weights`.

        Now, only accepts weight_scheme="average" or comma-separate numbers.

        Args:
            ensemble_weights: The passed-in argument.

        Returns: A list of floats. The size is the length of `self._model`.
        """
        if hasattr(self._model, "model_num"):
            model_num = self._model.model_num
        else:
            return None
        if ensemble_weights == "average":
            return [1. / float(model_num)] * model_num
        else:
            eles = ensemble_weights.strip().split(",")
            if len(eles) != model_num:
                raise ValueError("The number of manual weights must have equal to the models.")
            eles = [float(x) for x in eles]
            eles_sum = sum(eles)
            return [x / eles_sum for x in eles]

    def call(self, parsed_inp, **kwargs):
        """ Do beam search.

        Args:
            parsed_inp: A dict of parsed model inputs.

        Returns:
            The search results (must be a tuple).
        """
        maximum_decode_length = self._maximum_decode_length
        max_data_len = kwargs.get("max_data_len", None)
        if maximum_decode_length is None:
            assert max_data_len, ("`maximum_decode_length` must be provided "
                                  "when `max_data_len` is not provided.")
            maximum_decode_length = self._extra_decode_length + max_data_len
        if self._minimum_decode_length >= maximum_decode_length:
            raise ValueError("`minimum_decode_length` must be less than maximum decode length.")

        with tf.name_scope("beam_search"):
            decode_padded_length = maximum_decode_length if self._padded_decode else None
            symbols_to_logits_fn, generation_initializer = self._model.get_symbols_to_logits_fn(
                parsed_inp, is_training=False, is_inference=True,
                decode_padded_length=decode_padded_length)
        sorted_hypothesis, sorted_scores = sequence_beam_search(
            symbols_to_logits_fn,
            generation_initializer,
            top_k=self._top_k,
            beam_size=self._beam_size,
            length_penalty=self._length_penalty,
            extra_decode_length=self._extra_decode_length,
            maximum_decode_length=maximum_decode_length,
            minimum_decode_length=self._minimum_decode_length,
            ensemble_weights=self._get_ensemble_weights(self._ensemble_weights),
            padded_decode=self._padded_decode,
            enable_unk=self._enable_unk)
        return sorted_hypothesis, sorted_scores



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from neurst.utils.configurable import extract_constructor_params


class WordEmbeddingSharedWeights(tf.keras.layers.Layer):
    """Calculates input embeddings and pre-softmax linear with shared weights. """

    def __init__(self,
                 embedding_dim,
                 vocab_size,
                 share_softmax_weights=False,
                 use_bias=True,
                 verbose=False,
                 name=None):
        """ Initializes simple word embedding layer.

        Args:
            embedding_dim: An int scalar, the embedding dimension.
            vocab_size: An int scalar, the size of vocabulary.
            share_softmax_weights: A boolean, whether to share
                embedding table with target softmax weight.
            use_bias: A boolean, whether to use bias with target
                softmax weight.
            verbose: A boolean, whether to logging the parameters.
            name: The name of the layer.
        """
        self._params = extract_constructor_params(locals(), verbose=False)
        super(WordEmbeddingSharedWeights, self).__init__(name=name)
        self._embedding_dim = embedding_dim
        self._vocab_size = vocab_size
        self._share_softmax_weights = share_softmax_weights
        self._use_bias = use_bias

    @property
    def embedding_dim(self):
        return self._embedding_dim

    @property
    def vocab_size(self):
        return self._vocab_size

    def get_config(self):
        return self._params

    def build(self, input_shape):
        """ Builds the embedding table and softmax bias
            if share_softmax_weights=True."""
        if self._share_softmax_weights:
            scope_name = "shared"
        else:
            scope_name = "emb"
        with tf.name_scope(scope_name):
            self._shared_weights = self.add_weight(
                "weights",
                shape=(self._vocab_size, self._embedding_dim),
                trainable=True,
                initializer=tf.random_normal_initializer(
                    mean=0., stddev=self._embedding_dim ** -0.5))
            self._bias = None
            if self._share_softmax_weights and self._use_bias:
                self._bias = self.add_weight(
                    "bias",
                    shape=(self._vocab_size,),
                    # initializer=tf.zeros_initializer,
                    trainable=True)
        super(WordEmbeddingSharedWeights, self).__init__(input_shape)

    def _bottom(self, x):
        """ Embedding lookup.

        Args:
            x: A 1/2-d Tensor to be embedded.

        Returns: A 2/3-d Tensor according to `x`.
        """
        emb = tf.gather(self._shared_weights, x)
        return emb

    def _top(self, x):
        """ Computes logits on the top layer.

        Args:
            x: A Tensor with shape [..., hidden]

        Returns: A logits Tensor with shape [..., vocab_size].
        """
        original_shape = tf.shape(x)
        logits = tf.matmul(tf.reshape(x, [-1, self._embedding_dim]),
                           self._shared_weights, transpose_b=True)
        if self._bias is not None:
            logits += self._bias

        return tf.reshape(logits, tf.concat(
            [original_shape[:-1], [self._vocab_size]], axis=0))

    def call(self, inputs, mode="embedding", **kwargs):
        """ Gets token em beddings or computes logits.

        Args:
            inputs: An int tensor with shape [batch_size, length] or [batch, ].
            mode: A string, a valid value is one of "embedding" and "linear".

        Returns:
            A float tensor with shape [batch, length, embedding_dim]
            or [batch, embedding_dim] when mode == "embedding" ;
            A float tensor with shape [batch, length, vocab_size]
            when mode == "linear".
        """
        _ = kwargs
        if mode == "embedding":
            return self._bottom(inputs)
        elif mode == "linear":
            return self._top(inputs)
        else:
            raise ValueError("mode = {} is not valid.".format(mode))


class TokenTypeEmbedding(tf.keras.layers.Layer):
    """Calculates input embeddings and pre-softmax linear with shared weights. """

    def __init__(self,
                 embedding_dim,
                 vocab_size,
                 verbose=False,
                 name=None):
        """ Initializes simple word embedding layer.

        Args:
            embedding_dim: An int scalar, the embedding dimension.
            vocab_size: An int scalar, the size of vocabulary.
            share_softmax_weights: A boolean, whether to share
                embedding table with target softmax weight.
            verbose: A boolean, whether to logging the parameters.
            name: The name of the layer.
        """
        self._params = extract_constructor_params(locals(), verbose=False)
        super(TokenTypeEmbedding, self).__init__(name=name)
        self._embedding_dim = embedding_dim
        self._vocab_size = vocab_size
        self._emb_table = None

    @property
    def embedding_dim(self):
        return self._embedding_dim

    @property
    def vocab_size(self):
        return self._vocab_size

    def get_config(self):
        return self._params

    def build(self, input_shape):
        with tf.name_scope("token_type_emb"):
            self._emb_table = self.add_weight(
                "weights",
                shape=(self._vocab_size, self._embedding_dim),
                trainable=True,
                initializer=tf.random_normal_initializer(
                    mean=0., stddev=self._embedding_dim ** -0.5))
        super(TokenTypeEmbedding, self).__init__(input_shape)

    def call(self, inputs, **kwargs):
        """ Gets token em beddings or computes logits.

        Args:
            inputs: An int tensor with shape [batch_size, length] or [batch, ].

        Returns:
            A float tensor with shape [batch, length, embedding_dim] or [batch, embedding_dim].
        """
        _ = kwargs
        # This vocab will be small so we always do one-hot here, since it is always
        # faster for a small vocabulary.
        flat_token_type_ids = tf.reshape(inputs, [-1])
        one_hot_ids = tf.one_hot(flat_token_type_ids, depth=self.vocab_size)
        token_type_embeddings = tf.matmul(tf.cast(one_hot_ids, self._emb_table.dtype),
                                          self._emb_table)
        token_type_embeddings = tf.reshape(token_type_embeddings,
                                           tf.concat([tf.shape(inputs), [self.embedding_dim]], axis=0))
        return token_type_embeddings


class BertEmbedding(tf.keras.layers.Layer):
    """ The bert embedding layer. """

    def __init__(self,
                 embedding_dim,
                 vocab_size,
                 max_positions,
                 token_types,
                 dropout_rate=0.0,
                 epsilon=1e-12,
                 name=None):
        self._params = extract_constructor_params(locals(), verbose=False)
        super(BertEmbedding, self).__init__(name=name)
        self._embedding_dim = embedding_dim
        self._vocab_size = vocab_size
        self._max_positions = max_positions
        self._token_types = token_types
        self._dropout_rate = dropout_rate
        self._epsilon = epsilon

    @property
    def embedding_dim(self):
        return self._embedding_dim

    @property
    def vocab_size(self):
        return self._vocab_size

    @property
    def token_types(self):
        return self._token_types

    def get_config(self):
        return self._params

    def build(self, input_shape):
        self._word_embedding = self.add_weight(
            "word_embedding",
            shape=(self._vocab_size, self._embedding_dim),
            trainable=True,
            initializer=tf.random_normal_initializer(
                mean=0., stddev=self._embedding_dim ** -0.5))
        self._token_type_embedding = self.add_weight(
            "token_type_embedding",
            shape=(self._token_types, self._embedding_dim),
            trainable=True,
            initializer=tf.random_normal_initializer(
                mean=0., stddev=self._embedding_dim ** -0.5))
        self._position_embedding = self.add_weight(
            "position_embedding",
            shape=(self._max_positions, self._embedding_dim),
            initializer=tf.random_normal_initializer(
                mean=0., stddev=self._embedding_dim ** -0.5),
            trainable=True)
        self._norm_layer = tf.keras.layers.LayerNormalization(
            epsilon=self._epsilon, dtype="float32", name="ln")
        super(BertEmbedding, self).__init__(input_shape)

    def call(self, tokens, tokens_type=None, is_training=False):
        """ Gets bert embedding.

        Args:
            tokens: An int tensor with shape [batch_size, length].
            tokens_type: An int tensor of shape [batch_size, length].

        Returns:
            A float tensor with shape [batch, length, embedding_dim].
        """
        assert tokens.get_shape().ndims == 2
        embeddings = tf.gather(self._word_embedding, tokens)
        # This vocab will be small so we always do one-hot here, since it is always
        # faster for a small vocabulary.
        if tokens_type is None:
            tokens_type = tf.zeros_like(tokens, dtype=tf.int64)
        flat_token_type_ids = tf.reshape(tokens_type, [-1])
        one_hot_ids = tf.one_hot(flat_token_type_ids, depth=self.token_types)
        token_type_embeddings = tf.matmul(tf.cast(one_hot_ids, self._token_type_embedding.dtype),
                                          self._token_type_embedding)
        token_type_embeddings = tf.reshape(token_type_embeddings,
                                           tf.concat([tf.shape(tokens_type), [self.embedding_dim]], axis=0))
        position_embeddings = tf.expand_dims(tf.slice(self._position_embedding, [0, 0],
                                                      [tf.shape(tokens)[1], -1]), axis=0)
        bert_emb = self._norm_layer(embeddings + token_type_embeddings + position_embeddings)
        if is_training:
            bert_emb = tf.nn.dropout(bert_emb, rate=self._dropout_rate)
        return bert_emb



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from abc import ABCMeta, abstractmethod

import six
from absl import logging


@six.add_metaclass(ABCMeta)
class Tokenizer(object):
    """ Base class for text tokenizer. """

    def __init__(self, language, glossaries=None, **kwargs):
        """ Initializes the tokenizer. """
        _ = kwargs
        self._language = language
        self._glossaries = glossaries
        if self._glossaries is None:
            self._glossaries = []
        if len(self._glossaries) > 0:
            logging.info("WARNING: note that not all tokenizers support glossaries.")

    @property
    def language(self):
        """ The language. """
        return self._language

    def init_subtokenizer(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def tokenize(self, text, return_str=False):
        """ Tokenizes the text. """
        raise NotImplementedError

    @abstractmethod
    def detokenize(self, text, return_str=True):
        """ Detokenizes the text. """
        raise NotImplementedError

    @staticmethod
    def _convert_to_str(text, delimiter=" "):
        """ Converts the `text` to a string.

        Args:
            text: A string or a list of string.
            delimiter: The delimater for join.

        Returns: A string.
        """
        if isinstance(text, str):
            return text
        elif isinstance(text, list):
            return delimiter.join(text)
        else:
            raise ValueError("Not supported type of text: {}".format(type(text)))

    @staticmethod
    def _convert_to_list(text):
        """ Converts the `text` to a list of string.

        Args:
            text: A string or a list of string.

        Returns: A list of string.
        """
        if isinstance(text, str):
            return text.strip().split()
        elif isinstance(text, list):
            return text
        else:
            raise ValueError("Not supported type of text: {}".format(type(text)))

    @staticmethod
    def _output_wrapper(text, return_str, delimiter=" "):
        """ Converts the output `text` to string if `return_str` is True,
            else converters it to a list.

        Args:
            text: A string or a list of string.
            return_str: A bool, whether to convert to string.
            delimiter: The delimiter string for join.

        Returns:
            A string or a list of string according to `return_str`.
        """
        if isinstance(text, str):
            if return_str:
                return text
            else:
                return text.split()
        elif isinstance(text, list):
            if return_str:
                return delimiter.join(text)
            else:
                return text
        else:
            raise ValueError("Not supported type of text: {}".format(type(text)))

    @staticmethod
    def cjk_deseg(text):
        """ Desegment function for Chinese, Japanese and Korean.

        Args:
            text: A string.

        Returns:
            The desegmented string.
        """

        def _strip(matched):
            return matched.group(1).strip()

        char_space_pattern1 = r"([\u2E80-\u9FFF\uA000-\uA4FF\uAC00-\uD7FF\uF900-\uFAFF]\s+)"
        char_space_pattern2 = r"(\s+[\u2E80-\u9FFF\uA000-\uA4FF\uAC00-\uD7FF\uF900-\uFAFF])"

        res = re.sub(char_space_pattern1, _strip, text)
        res = re.sub(char_space_pattern2, _strip, res)
        # no leading space
        res = re.sub(r'^\s+', r'', res)

        # no trailing space
        res = re.sub(r'\s+$', r'', res)
        return res



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from sacremoses.normalize import MosesPunctNormalizer

from neurst.data.data_pipelines import register_data_pipeline
from neurst.data.data_pipelines.text_data_pipeline import TextDataPipeline

PUNC_PATTERN = re.compile(r"[,\.\!\(\);:、\?\-\+=\"/><《》\[\]，。：；「」【】{}`@#\$%\^&\*]")


@register_data_pipeline("transcript")
class TranscriptDataPipeline(TextDataPipeline):
    PUNC_NORMERS = dict()

    def __init__(self, *args, **kwargs):
        """ Initializes the data pipeline for transcription(audio2text) data.

        Args:
            remove_punctuation: Whether to remove punctuation.
            lowercase: Whether to lowercase the text.
        """
        self._remove_punc = kwargs.get("remove_punctuation", False)
        self._lc = kwargs.get("lowercase", "True")
        super(TranscriptDataPipeline, self).__init__(*args, **kwargs)

    @staticmethod
    def cleanup_transcript(language, transcript, lowercase=True, remove_punctuation=True):
        if lowercase:
            transcript = transcript.lower()
        if language not in ["zh", "ja"]:
            if language not in TranscriptDataPipeline.PUNC_NORMERS:
                TranscriptDataPipeline.PUNC_NORMERS[language] = MosesPunctNormalizer(lang=language)
            transcript = TranscriptDataPipeline.PUNC_NORMERS[language].normalize(transcript)
            transcript = transcript.replace("' s ", "'s ").replace(
                "' ve ", "'ve ").replace("' m ", "'m ").replace("' t ", "'t ").replace("' re ", "'re ")
        if remove_punctuation:
            transcript = PUNC_PATTERN.sub(" ", transcript)
        transcript = " ".join(transcript.strip().split())
        return transcript

    def process(self, input, is_processed=False):
        if not is_processed:
            input = self.cleanup_transcript(self._language, input, self._lc, self._remove_punc)
        return super(TranscriptDataPipeline, self).process(input, is_processed)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Defines utils for distributed training. """
import json
import os

import tensorflow as tf

from neurst.utils.compat import register_distributed_worker_setting


def tpu_initialize(tpu_address):
    """Initializes TPU for TF 2.0 training.

    Args:
      tpu_address: string, bns address of master TPU worker.

    Returns:
      A TPUClusterResolver.
    """
    cluster_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(
        tpu=tpu_address)
    if tpu_address not in ('', 'local'):
        tf.config.experimental_connect_to_cluster(cluster_resolver)
    tf.tpu.experimental.initialize_tpu_system(cluster_resolver)
    return cluster_resolver


def _collective_communication(all_reduce_alg):
    """Return a CollectiveCommunication based on all_reduce_alg.

    Args:
      all_reduce_alg: a string specifying which collective communication to pick,
        or None.

    Returns:
      tf.distribute.experimental.CollectiveCommunication object

    Raises:
      ValueError: if `all_reduce_alg` not in [None, 'ring', 'nccl']
    """
    collective_communication_options = {
        None: tf.distribute.experimental.CollectiveCommunication.AUTO,
        "ring": tf.distribute.experimental.CollectiveCommunication.RING,
        "nccl": tf.distribute.experimental.CollectiveCommunication.NCCL
    }
    if all_reduce_alg not in collective_communication_options:
        raise ValueError(
            "When used with `multi_worker_mirrored`, valid values for "
            "all_reduce_alg are ['ring', 'nccl'].  Supplied value: {}".format(
                all_reduce_alg))
    return collective_communication_options[all_reduce_alg]


def _mirrored_cross_device_ops(all_reduce_alg, num_packs):
    """Return a CrossDeviceOps based on all_reduce_alg and num_packs.

    Args:
      all_reduce_alg: a string specifying which cross device op to pick, or None.
      num_packs: an integer specifying number of packs for the cross device op.

    Returns:
      tf.distribute.CrossDeviceOps object or None.

    Raises:
      ValueError: if `all_reduce_alg` not in [None, 'nccl', 'hierarchical_copy'].
    """
    if all_reduce_alg is None:
        return None
    mirrored_all_reduce_options = {
        "nccl": tf.distribute.NcclAllReduce,
        "hierarchical_copy": tf.distribute.HierarchicalCopyAllReduce
    }
    if all_reduce_alg not in mirrored_all_reduce_options:
        raise ValueError(
            "When used with `mirrored`, valid values for all_reduce_alg are "
            "['nccl', 'hierarchical_copy'].  Supplied value: {}".format(
                all_reduce_alg))
    cross_device_ops_class = mirrored_all_reduce_options[all_reduce_alg]
    return cross_device_ops_class(num_packs=num_packs)


def get_distribution_strategy(distribution_strategy="mirrored",
                              num_gpus=0,
                              worker_hosts=None,
                              task_index=-1,
                              all_reduce_alg="nccl",
                              num_packs=1,
                              tpu_address=None):
    """Return a DistributionStrategy for running the model.

    Args:
      distribution_strategy: a string specifying which distribution strategy to
        use. Accepted values are 'off', 'default', 'one_device', 'mirrored',
        'parameter_server', 'multi_worker_mirrored', and 'tpu' -- case insensitive.
        'off' means not to use Distribution Strategy; 'default' means to choose from
        `MirroredStrategy`, `MultiWorkerMirroredStrategy`, or `OneDeviceStrategy`
        according to the number of GPUs and number of workers. 'tpu' means to use
        TPUStrategy using `tpu_address`.
      num_gpus: Number of GPUs to run this model.
      worker_hosts: The worker hosts for 'multi_worker_mirrored'.
      task_index: The task index for 'multi_worker_mirrored'.
      all_reduce_alg: Optional. Specifies which algorithm to use when performing
        all-reduce. For `MirroredStrategy`, valid values are "nccl" and
        "hierarchical_copy". For `MultiWorkerMirroredStrategy`, valid values are
        "ring" and "nccl".  If None, DistributionStrategy will choose based on
        device topology.
      num_packs: Optional.  Sets the `num_packs` in `tf.distribute.NcclAllReduce`
        or `tf.distribute.HierarchicalCopyAllReduce` for `MirroredStrategy`.
      tpu_address: Optional. String that represents TPU to connect to. Must not
        be None if `distribution_strategy` is set to `tpu`.
    Returns:
      tf.distribute.DistibutionStrategy object.
    Raises:
      ValueError: if `distribution_strategy` is 'off' or 'one_device' and
        `num_gpus` is larger than 1; or `num_gpus` is negative or if
        `distribution_strategy` is `tpu` but `tpu_address` is not specified.
    """
    if num_gpus == 0:
        num_gpus = int(os.environ.get("WORKER_GPUS", '0'))
    if num_gpus < 0:
        raise ValueError("`num_gpus` can not be negative.")
    if (distribution_strategy is None or distribution_strategy.lower() == "none"
        or distribution_strategy == ""):
        return None
    distribution_strategy = distribution_strategy.lower()

    if distribution_strategy == "tpu":
        # When tpu_address is an empty string, we communicate with local TPUs.
        cluster_resolver = tpu_initialize(tpu_address)
        return tf.distribute.experimental.TPUStrategy(cluster_resolver)

    if distribution_strategy == "multi_worker_mirrored":
        if worker_hosts is None:
            worker_hosts = os.environ.get("WORKER_HOSTS", None)
            task_index = int(os.environ.get("TASK_ID", -1))
        assert worker_hosts, (
            "worker_hosts must be provided when using 'multi_worker_mirrored'.")

        workers = worker_hosts.split(',')
        if len(workers) > 1 and 0 > task_index:
            raise ValueError('Must specify task_index when number of workers > 1')
        task_index = 0 if len(workers) == 1 else task_index
        register_distributed_worker_setting(worker_id=task_index, num_workers=len(workers),
                                            strategy="multi_worker_mirrored")
        os.environ['TF_CONFIG'] = json.dumps({
            'cluster': {
                'worker': workers
            },
            'task': {'type': 'worker', 'index': task_index}
        })
        strategy = tf.distribute.experimental.MultiWorkerMirroredStrategy(
            communication=_collective_communication(all_reduce_alg))
        strategy.extended.experimental_enable_get_next_as_optional = True
        return strategy

    if distribution_strategy in ("mirrored", "default"):
        return tf.distribute.MirroredStrategy(
            cross_device_ops=_mirrored_cross_device_ops(
                all_reduce_alg, num_packs))

    raise ValueError(
        "Unrecognized Distribution Strategy: %r" % distribution_strategy)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

import tensorflow as tf
from absl import logging

from neurst.criterions import Criterion, build_criterion
from neurst.data.datasets import Dataset, build_dataset
from neurst.data.datasets.multiple_dataset import MultipleDataset
from neurst.training import Validator, register_validator, training_utils
from neurst.utils import compat
from neurst.utils.flags_core import Flag, ModuleFlag


@register_validator
class CriterionValidator(Validator):

    def __init__(self, args):
        super(CriterionValidator, self).__init__(args)
        self.args = args  # lazy initialization
        self._criterion = None
        self._criterion_metric = None
        self._custom_dataset = None
        self._strategy = None
        self._eval_tfds = None
        self._criterion_model = None
        self._criterion_recorder = None
        self._avg_criterion_recorder = None
        self._mixed_criterion_recorder = None
        self._criterion_start_time = None
        self._validate_criterion = True
        self._eval_task_args = args["eval_task_args"] or {}
        self._eval_task_args["batch_size"] = args["eval_batch_size"]

    @staticmethod
    def class_or_method_args():
        this_args = super(CriterionValidator, CriterionValidator).class_or_method_args()
        this_args.extend([
            ModuleFlag("eval_criterion", Criterion.REGISTRY_NAME, help="The criterion for validation."),
            ModuleFlag("eval_dataset", Dataset.REGISTRY_NAME, help="The dataset for validation."),
            Flag("eval_batch_size", dtype=Flag.TYPE.INTEGER, default=32,
                 help="The batch size for validation process."),
            Flag("eval_task_args", dtype=Flag.TYPE.INTEGER, default=None,
                 help="Other parameters for building validation dataset.")
        ])
        return this_args

    def build(self, strategy, task, model):
        """ Initializes. """
        self._strategy = strategy
        self._criterion: Criterion = build_criterion(self.args["eval_criterion.class"],
                                                     **self.args["eval_criterion.params"])
        self._criterion.set_model(model)
        if self._criterion is None:
            logging.info("WARNING: no criterion is provided in CriterionValidator "
                         "for validation process.")
            self._validate_criterion = False
            return self
        self._custom_dataset = build_dataset(self.args["eval_dataset.class"], **self.args["eval_dataset.params"])
        if self._custom_dataset is None:
            logging.info("WARNING: no validation dataset is provided "
                         "in CriterionValidator for validation process.")
            self._validate_criterion = False
            return self
        from neurst.exps.evaluator import Evaluator
        with training_utils.get_strategy_scope(strategy):
            self._criterion_model = Evaluator.build_evaluation_model(task, model, self._criterion)
            self._eval_tfds = training_utils.build_datasets(
                compat.ModeKeys.EVAL, strategy, self._custom_dataset, task, self._eval_task_args)
        self._criterion_metric = self._criterion.as_metric()
        if isinstance(self._custom_dataset, MultipleDataset):
            self._criterion_recorder = {
                name: training_utils.TrainingStatusRecorder(
                    model=model, task=task, metric=self._criterion_metric)
                for name in self._custom_dataset.datasets
            }
            self._avg_criterion_recorder = training_utils.TrainingStatusRecorder(
                model=model, task=task, metric=self._criterion_metric)
            self._mixed_criterion_recorder = training_utils.TrainingStatusRecorder(
                model=model, task=task, metric=self._criterion_metric)
        else:
            self._criterion_recorder = training_utils.TrainingStatusRecorder(
                model=model, task=task, metric=self._criterion_metric)
        self._criterion_start_time = time.time()
        return self

    def validate(self, step):
        if not self._validate_criterion:
            return
        start_time = time.time()
        results, avg_res, mixed_res = training_utils.reduce_eval_results(
            self._criterion, self._custom_dataset, training_utils.make_predictions(
                self._strategy, self._criterion_model, self._eval_tfds, self._custom_dataset))
        elapsed = time.time() - start_time
        elapsed_from_start = time.time() - self._criterion_start_time

        def _display(res, best, name=None, tb_name=None):
            if tb_name is None:
                tb_name = name
            tb_name = "" if tb_name is None else (tb_name + "_")
            name = "" if name is None else f" ({name})"
            for k, v in res.items():
                logging.info("Evaluating (%s) validation set%s: %s=%.2f (Best %.2f)  "
                             "step=%d\tElapsed %.2fs  FromSTART %.2fs",
                             self._criterion_metric.flag, name, k, v, best[k],
                             step, elapsed, elapsed_from_start)
                tf.summary.scalar(
                    compat.GlobalKeys.TBPREFIX_VALID + f"/{tb_name}{k}", v, step=step)

        if isinstance(self._custom_dataset, MultipleDataset):
            for name, res in results.items():
                self._criterion_recorder[name].record(step, res)
                _display(res, self._criterion_recorder[name].best, name=name)
            self._avg_criterion_recorder.record(step, avg_res)
            _display(avg_res, self._avg_criterion_recorder.best,
                     f"on average by weights {self._custom_dataset.sample_weights}",
                     tb_name="AVERAGE")
            self._mixed_criterion_recorder.record(step, mixed_res)
            _display(mixed_res, self._mixed_criterion_recorder.best, "MIXED")
        else:
            self._criterion_recorder.record(step, results)
            _display(results, self._criterion_recorder.best)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Base Decoder class. """
from __future__ import absolute_import, division, print_function

from abc import ABCMeta, abstractmethod

import six
import tensorflow as tf

from neurst.utils.configurable import extract_constructor_params


@six.add_metaclass(ABCMeta)
class Decoder(tf.keras.layers.Layer):
    """Base class for decoders. """
    REGISTRY_NAME = "decoder"

    def __init__(self, name=None, **kwargs):
        """ Initializes the parameters of the decoder. """
        self._params = extract_constructor_params(locals(), verbose=False)
        super(Decoder, self).__init__(name=name)

    def build(self, input_shape):
        super(Decoder, self).build(input_shape)

    @abstractmethod
    def create_decoding_internal_cache(self,
                                       encoder_outputs,
                                       encoder_inputs_padding,
                                       is_inference=False,
                                       decode_padded_length=None):
        """ Creates internal cache for decoding.

        Args:
            encoder_outputs: The output tensor from encoder
                with shape [batch_size, max_input_length, hidden_size].
            encoder_inputs_padding: A float tensor with shape [batch_size, max_length],
                indicating the padding positions of `encoder_output`, where 1.0 for
                padding and 0.0 for non-padding.
            is_inference: A boolean scalar, whether in inference mode or not.
            decode_padded_length: The maximum decoding length when inference, for creating
                static-shape cache.

        Returns:
            `cache`, a dictionary containing static(e.g. encoder hidden states
            for attention) and dynamic(e.g. transformer decoding cache) tensors used
            during decoding and will be passed to `call()`. Note that, the dynamic
            tensors must store in cache["decoding_states"] for beam search use.
        """
        raise NotImplementedError

    def get_config(self):
        return self._params

    @abstractmethod
    def call(self, decoder_inputs, cache, is_training=True, decode_loop_step=None):
        """ Encodes the inputs.

        Args:
            decoder_inputs: The embedded decoder input, a float tensor with shape
                [batch_size, max_target_length, embedding_dim] or
                [batch_size, embedding_dim] for one decoding step.
            cache: A dictionary, generated from self.create_decoding_internal_cache.
            is_training: A bool, whether in training mode or not.
            decode_loop_step: An integer, step number of the decoding loop. Used only
                for autoregressive inference with static-shape cache.

        Returns:
            The decoder output with shape [batch_size, max_length, hidden_size]
            when `decoder_inputs` is a 3-d tensor or with shape
            [batch_size, hidden_size] when `decoder_inputs` is a 2-d tensor.
        """
        raise NotImplementedError



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re

import numpy
import tensorflow as tf
import yaml
from absl import logging

from neurst.models.gpt2 import GPT2
from neurst.utils.converters import Converter, register_converter
from neurst.utils.misc import download_with_tqdm

_URL_PREFIX = "https://storage.googleapis.com/gpt-2"
_GPT2_PRETRAIN_MODELS = {
    "117M": _URL_PREFIX + "/models/117M",
    "345M": _URL_PREFIX + "/models/345M",
}

_DIRECT_MAPPINGS = {
    "model/wte": "gpt2/posenc_wrapper/embeddings/shared/weights",
    "model/wpe": "gpt2/posenc_wrapper/position_embeddings/weights",
    "model/ln_f/b": "gpt2/decoder/output_ln/beta",
    "model/ln_f/g": "gpt2/decoder/output_ln/gamma",
}

_POSTFIX_MAPPINGS = {
    "attn/c_attn/w": "self_attention_prepost_wrapper/self_attention/qkv_transform/kernel",
    "attn/c_attn/b": "self_attention_prepost_wrapper/self_attention/qkv_transform/bias",
    "attn/c_proj/w": "self_attention_prepost_wrapper/self_attention/output_transform/kernel",
    "attn/c_proj/b": "self_attention_prepost_wrapper/self_attention/output_transform/bias",
    "ln_1/b": "self_attention_prepost_wrapper/ln/beta",
    "ln_1/g": "self_attention_prepost_wrapper/ln/gamma",
    "ln_2/b": "ffn_prepost_wrapper/ln/beta",
    "ln_2/g": "ffn_prepost_wrapper/ln/gamma",
    "mlp/c_fc/w": "ffn_prepost_wrapper/ffn/dense1/kernel",
    "mlp/c_fc/b": "ffn_prepost_wrapper/ffn/dense1/bias",
    "mlp/c_proj/w": "ffn_prepost_wrapper/ffn/dense2/kernel",
    "mlp/c_proj/b": "ffn_prepost_wrapper/ffn/dense2/bias"
}


@register_converter(["openai_gpt2", "gpt2"])
class OpenAIGPT2(Converter):
    """ Converts from the from openai gpt-2.
    https://github.com/openai/gpt-2
    """

    @staticmethod
    def convert_model_config(path):
        with tf.io.gfile.GFile(os.path.join(path, "hparams.json")) as fp:
            cfg = yaml.load(fp, Loader=yaml.FullLoader)
        return {
            "model.class": GPT2.__name__,
            "model.params": {
                "max_position_embeddings": cfg["n_ctx"],
                "num_layers": cfg["n_layer"],
                "hidden_size": cfg["n_embd"],
                "num_attention_heads": cfg["n_head"]
            }
        }

    @staticmethod
    def convert_task_config(path):
        raise NotImplementedError

    @staticmethod
    def download(key):
        if key in _GPT2_PRETRAIN_MODELS:
            url = _GPT2_PRETRAIN_MODELS[key]
        elif key.startswith("http://") or key.startswith("https://"):
            url = key
        else:
            return None
        logging.info(f"Downloading openai gpt2: {key}")
        this_dir = os.path.dirname(__file__)
        model_dir = os.path.join(this_dir, f"GPT2_{key}")
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)
        for filename in ['checkpoint', 'encoder.json', 'hparams.json', 'model.ckpt.data-00000-of-00001',
                         'model.ckpt.index', 'model.ckpt.meta', 'vocab.bpe']:
            this_url = url + "/" + filename
            save_filename = os.path.join(model_dir, filename)
            if not os.path.exists(save_filename):
                logging.info(f"Downloading {this_url}")
                download_with_tqdm(this_url, save_filename)
        return model_dir

    @staticmethod
    def convert_checkpoint(path, save_path):
        ckpt = os.path.join(path, "model.ckpt")
        gpt2_var_names = [x[0] for x in tf.train.list_variables(ckpt)]
        new_vars = []
        processed_var_names = []
        for var_name in gpt2_var_names:
            var_value = numpy.squeeze(tf.train.load_variable(ckpt, var_name))
            new_var_name = None
            if var_name in _DIRECT_MAPPINGS:
                new_var_name = _DIRECT_MAPPINGS[var_name]
            elif var_name.startswith("model/h"):
                lid = re.search(r"h\d+", var_name).group()[1:]
                postfix = var_name.split(f"h{lid}/")[1]
                if postfix in _POSTFIX_MAPPINGS:
                    new_var_name = f"gpt2/decoder/layer_{lid}/{_POSTFIX_MAPPINGS[postfix]}"
                else:
                    raise Exception("wrong")

            if new_var_name:
                processed_var_names.append(new_var_name)
                logging.info(f"convert {var_name}")
                logging.info(f"\t ==> {new_var_name}")
                new_vars.append(tf.Variable(
                    initial_value=var_value,
                    trainable=False,
                    name=new_var_name,
                    dtype=str(var_value.dtype)))
            else:
                logging.info(f"No matching variable for {var_name}")
        ckpt_saver = tf.train.Checkpoint(**{x.name.split(":")[0]: x for x in new_vars})
        save_ckpt = os.path.join(save_path, "ckpt")
        logging.info(f"Saving checkpoint to {save_ckpt}")
        ckpt_saver.save(save_ckpt)



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABCMeta, abstractmethod

import six
import tensorflow as tf
from absl import logging

from neurst.data.datasets import register_dataset
from neurst.data.datasets.dataset import TFRecordDataset
from neurst.data.datasets.text_gen_dataset import TextGenDataset
from neurst.utils.compat import DataStatus
from neurst.utils.flags_core import Flag


@six.add_metaclass(ABCMeta)
class AbstractParallelDataset(TextGenDataset):
    """ The abstract dataset for parallel text.
    The element spec must be
        {
            'feature': tf.TensorSpec(shape=(None,), dtype=tf.int64),
            'label': tf.TensorSpec(shape=(None,), dtype=tf.int64)
         }
    """

    def __init__(self):
        self._sources = None
        super(AbstractParallelDataset, self).__init__()

    @property
    @abstractmethod
    def status(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def build_iterator(self, map_func=None, shard_id=0, total_shards=1):
        """ Returns the iterator of the dataset.

        Args:
            map_func: A function mapping a dataset element to another dataset element.
            shard_id: Generator yields on the `shard_id`-th shard of the whole dataset.
            total_shards: The number of total shards.
        """
        raise NotImplementedError

    @property
    def sources(self):
        """ Returns a list of source texts. """
        return self._sources


@register_dataset("parallel_text")
class ParallelTextDataset(AbstractParallelDataset):

    def __init__(self, args):
        """ Initializes the dataset. """
        super(ParallelTextDataset, self).__init__()
        self._src_file = args["src_file"]
        assert self._src_file, "`src_file` must be provided for ParallelTextDataset."
        self._trg_file = args["trg_file"]
        self._data_is_processed = args["data_is_processed"]

    @staticmethod
    def class_or_method_args():
        return [
            Flag("src_file", dtype=Flag.TYPE.STRING, help="The source text file"),
            Flag("trg_file", dtype=Flag.TYPE.STRING, help="The target text file"),
            Flag("data_is_processed", dtype=Flag.TYPE.BOOLEAN,
                 help="Whether the text data is already processed."),
        ]

    @property
    def status(self):
        if self._data_is_processed:
            return DataStatus.PROCESSED
        return DataStatus.RAW

    def build_iterator(self, map_func=None, shard_id=0, total_shards=1):
        """ Reads data from files and returns the iterator.

        Args:
            map_func: A function mapping a dataset element to another dataset element.
            shard_id: Generator yields on the `shard_id`-th shard of the whole dataset.
            total_shards: The number of total shards.
        """
        if total_shards > 1:
            total_samples = self.num_samples
            samples_per_part = total_samples // total_shards
            range_begin = samples_per_part * shard_id
            if shard_id == total_shards - 1:
                range_end = total_samples + 1
                logging.info(f"Iterate on dataset from {range_begin} "
                             f"to the end (total {total_samples}).")
            else:
                range_end = range_begin + samples_per_part
                logging.info(f"Iterate on dataset from {range_begin} "
                             f"to {range_end} (total {total_samples}).")

        def gen():
            fsrc = tf.io.gfile.GFile(self._src_file)
            ftrg = None if self._trg_file is None else tf.io.gfile.GFile(self._trg_file)
            n = 0
            for src in fsrc:
                n += 1
                data = {"feature": src.strip()}
                if ftrg is not None:
                    data["label"] = ftrg.readline().strip()
                if total_shards > 1:
                    if n < range_begin:
                        continue
                    if n >= range_end:
                        break
                if map_func is not None:
                    data = map_func(data)
                yield data
            fsrc.close()
            if ftrg is not None:
                ftrg.close()

        return gen

    @property
    def sources(self):
        """ Returns a list of sources. """
        if self._sources is None and self._src_file:
            with tf.io.gfile.GFile(self._src_file) as fp:
                self._sources = [line.strip() for line in fp]
        return self._sources

    @property
    def targets(self):
        """ Returns a list of targets. """
        if self._targets is None and self._trg_file:
            with tf.io.gfile.GFile(self._trg_file) as fp:
                self._targets = [line.strip() for line in fp]
        return self._targets


@register_dataset("parallel_tfrecord")
class ParallelTFRecordDataset(TFRecordDataset, AbstractParallelDataset):

    @property
    def status(self):
        return DataStatus.PROJECTED

    @property
    def fields(self):
        return {"feature": tf.io.VarLenFeature(tf.int64),
                "label": tf.io.VarLenFeature(tf.int64)}



# Copyright 2020 ByteDance Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from absl import logging

from neurst.criterions import build_criterion, register_criterion
from neurst.criterions.criterion import Criterion
from neurst.utils.flags_core import Flag


@register_criterion
class JointCriterion(Criterion):

    def __init__(self, args):
        """ Initializes the cross entropy with label smoothing.

        Args:
            args: A dict of full parameters.
        """
        super(JointCriterion, self).__init__()
        assert isinstance(args["criterions"], list), (
            "`criterions` should be a list of multiple criterion settings for JointCriterion.")
        self._criterions = []
        self._alphas = []
        logging.info("Creating JointCriterion for training, which is composed by:")
        for crit in args["criterions"]:
            self._alphas.append(crit.pop("alpha", 1.0))
            self._criterions.append(build_criterion(crit))
            logging.info("  - {} with alpha={}".format(self._criterions[-1].__class__.__name__, self._alphas[-1]))

    @staticmethod
    def class_or_method_args():
        """ Returns a list of args for flag definition. """
        return [Flag("criterions", dtype=Flag.TYPE.STRING,
                     default=None, help="A list of multiple criterions. Each element should be a dict like "
                                        "\"{'criterion.class': '...', 'criterion.params': '...', 'alpha': 1.0}\","
                                        "where alpha denotes the weight of the corresponding criterion "
                                        "and will be set to 1.0 by default.")]

    def reduce_loss(self, model_inp, model_out):
        """ Reduces loss tensor for training according to the model inputs
            and outputs.

        Returns: A nest structure of tensors.
        """
        reduced_loss = dict()
        for alpha, criterion in zip(self._alphas, self._criterions):
            this_loss = criterion.reduce_loss(model_inp, model_out)
            # We now only consider to joint the criterion with single loss tensor.
            reduced_loss[criterion.__class__.__name__] = alpha * this_loss
        return reduced_loss

    def reduce_metrics(self, eval_res_list):
        """ Reduces the metrics according to a list of returned value from `eval`.

        Args:
            eval_res_list: A list of tuples of numpy.ndarray generated by `self.__call__`
                and model.__call__.

        Returns:
            A dict of reduced metrics for evaluation.
        """
        final_metric = dict()
        for criterion, eval_res_l in zip(self._criterions, list(map(list, zip(*eval_res_list)))):
            for k, v in criterion.reduce_metrics(eval_res_l).items():
                final_metric[criterion.__class__.__name__ + "/" + k] = v
        return final_metric

    def as_metric(self):
        """ Returns a wrapper class of Metric. """
        return self._criterions[0].as_metric()

    def __call__(self, model_inp, model_out):
        """ Returns a list of __call__ output corresponding to each criterion. """
        return [x(model_inp, model_out) for x in self._criterions]
